import pytest

from tackle import tackle
from tackle import exceptions


def test_parser_functions_raises_unknown_arg(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownArgumentException):
        tackle("cli-default-hook-no-context.yaml", 'NOT_HERE')


def test_parser_functions_raises_unknown_kwarg(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("cli-default-hook-no-context.yaml", NOT_HERE='foo')


def test_parser_functions_raises_unknown_flags(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("cli-default-hook-no-context.yaml", global_flags=['NOT_HERE'])


def test_parser_functions_raises_unknown_arg_hook(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownArgumentException):
        tackle("cli-hook-no-context.yaml", 'run', 'NOT_HERE')


def test_parser_functions_raises_unknown_kwarg_hook(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("cli-hook-no-context.yaml", 'run', NOT_HERE='foo')


def test_parser_functions_raises_unknown_flags_hook(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("cli-hook-no-context.yaml", 'run', global_flags=['NOT_HERE'])


def test_parser_functions_raises_hook_kwarg_missing(chdir):
    chdir('exceptions')
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("hook-kwarg-missing.yaml", 'foo', baz='bang')


def test_parser_functions_raises_hook_kwarg_missing_default(chdir):
    chdir('exceptions')
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("hook-kwarg-missing-default.yaml", baz='bang')



from collections import OrderedDict
from functools import partialmethod
import os
from pydantic import Field, create_model, ValidationError
from pydantic.main import ModelMetaclass
from pydantic.fields import ModelField
from pydoc import locate
from pathlib import Path
import re
from ruamel.yaml.constructor import CommentedKeyMap, CommentedMap
from ruamel.yaml.parser import ParserError
from typing import Type, Any, Union, Callable, Optional

from tackle import exceptions
from tackle.hooks import (
    import_from_path,
    import_with_fallback_install,
    LazyImportHook,
    LazyBaseFunction,
)
from tackle.macros import (
    var_hook_macro,
    blocks_macro,
    compact_hook_call_macro,
    list_to_var_macro,
    function_field_to_parseable_macro,
)
from tackle.models import (
    Context,
    BaseHook,
    BaseFunction,
    FunctionInput,
    BaseContext,
)
from tackle.render import render_variable
from tackle.settings import settings
from tackle.utils.dicts import (
    nested_get,
    nested_delete,
    nested_set,
    encode_list_index,
    decode_list_index,
    set_key,
    get_target_and_key,
    smush_key_path,
    get_readable_key_path,
    cleanup_unquoted_strings,
)
from tackle.utils.command import unpack_args_kwargs_string
from tackle.utils.files import read_config_file
from tackle.utils.help import run_help
from tackle.utils.imports import get_public_or_private_hook
from tackle.utils.render import wrap_jinja_braces
from tackle.utils.paths import (
    work_in,
    is_repo_url,
    is_directory_with_tackle,
    is_file,
    find_tackle_file,
    find_nearest_tackle_file,
    find_in_parent,
)
from tackle.utils.vcs import get_repo_source
from tackle.utils.zipfile import unzip

BASE_METHODS = [
    'if',
    'when',
    'else',
    'for',
    'reverse',
    'try',
    'except',
    'chdir',
    'merge',
]


def get_hook(hook_type, context: 'Context') -> Type[BaseHook]:
    """
    Get the hook from providers. Qualify if the hook is a method and if it is a lazy
    hook (ie has requirements that have not been installed), install them.
    """
    if '.' in hook_type:
        # When the hook type has a period in it, we are calling a hook's method. Here we
        # are going to split up the call into methods denoted by periods, and do logic
        # to support inheriting base properties into methods. To do this we need to
        # instantiate the base and use the `function_fields` list to inform which
        # fields will be inherited. See `create_function_model` function for more info.
        hook_parts = hook_type.split('.')

        # Extract the base hook.
        hook_type = hook_parts.pop(0)
        h = get_public_or_private_hook(context, hook_type)
        if h is None:
            exceptions.raise_unknown_hook(context, hook_type)

        # TODO: To support calling python hook methods, put a conditional here to
        #  determine if hook is of type BaseHook / LazyImportHook and perform logic
        #  separate from the declarative hook.
        for method in hook_parts:
            new_hook = None
            try:
                new_hook = h.__fields__[method].default
            except (IndexError, KeyError):
                # Raise error - method not found
                exceptions.raise_unknown_hook(context, method, method=True)

            # Update method with values from base class so that fields can be inherited
            # from the base hook. function_fields is a list of those fields that aren't
            # methods / special vars (ie args, return, etc).
            for i in h.__fields__['function_fields'].default:
                if i not in new_hook.function_dict:
                    # Base method should not override child.
                    new_hook.function_dict[i] = h.__fields__['function_dict'].default[i]

            # Methods that are of type LazyBaseFunction which need to have the base
            # instantiated before getting the hook. Allows nested methods for functions.
            if isinstance(new_hook, LazyBaseFunction):
                new_hook = create_function_model(
                    context=context,
                    func_name=h.__fields__[method].name,
                    func_dict=new_hook.function_dict.copy(),
                )

            h = new_hook

    else:
        h = get_public_or_private_hook(context, hook_type)
    if h is None:
        # Raise exception for unknown hook
        exceptions.raise_unknown_hook(context, hook_type)

    # LazyImportHook in hook ref when declared in provider __init__.hook_types
    elif isinstance(h, LazyImportHook):
        # Install the requirements which will convert all the hooks in that provider
        #  to actual hooks
        import_with_fallback_install(
            context=context,
            mod_name=h.mod_name,
            path=h.hooks_path,
        )
        h = get_public_or_private_hook(context, hook_type)
    # TODO: Refactor this whole function so this is not repeated
    #  Make it so hook is split right away and evaluated in one loop
    elif isinstance(h, LazyBaseFunction):
        h = create_function_model(
            context=context,
            func_name=hook_type,
            func_dict=h.function_dict.copy(),
        )

    return h


def evaluate_for(hook_dict: dict, Hook: ModelMetaclass, context: 'Context'):
    """Run the parse_hook function in a loop and return a list of outputs."""
    loop_targets = render_variable(context, wrap_jinja_braces(hook_dict['for']))

    if len(loop_targets) == 0:
        return

    hook_dict.pop('for')

    # Need add an empty list in the value so we have something to append to
    target_context, key_path = get_target_and_key(context)
    nested_set(target_context, key_path, [])

    # Account for nested contexts and justify the new keys based on the key path within
    #  blocks by trimming the key_path_block from the key_path.
    if len(context.key_path_block) != 0:
        tmp_key_path = context.key_path[
            len(context.key_path_block) - len(context.key_path) :
        ]  # noqa
        if context.temporary_context is None:
            context.temporary_context = {} if isinstance(tmp_key_path[0], str) else []
        tmp_key_path = [i for i in tmp_key_path if i not in ('->', '_>')]
        nested_set(context.temporary_context, tmp_key_path, [])

    for i, l in (
        enumerate(loop_targets)
        if not render_variable(context, hook_dict.get('reverse', None))
        else reversed(list(enumerate(loop_targets)))
    ):
        if context.existing_context is None:
            context.existing_context = {}
        # Create temporary variables in the context to be used in the loop.
        context.existing_context.update({'index': i, 'item': l})
        # Append the index to the keypath
        context.key_path.append(encode_list_index(i))

        # Reparse the hook with the new temp vars in place
        parse_hook(hook_dict.copy(), Hook, context, append_hook_value=True)
        context.key_path.pop()

    # Remove temp variables
    try:
        context.existing_context.pop('item')
        context.existing_context.pop('index')
    except KeyError:
        pass


def evaluate_if(hook_dict: dict, context: 'Context', append_hook_value: bool) -> bool:
    """Evaluate the if/when condition and return bool."""
    if hook_dict.get('when', None) is not None:
        result = render_variable(context, wrap_jinja_braces(hook_dict['when']))
        hook_dict.pop('when')
        return result
    if hook_dict.get('for', None) is not None and not append_hook_value:
        # We qualify `if` conditions within for loop logic
        return True
    if hook_dict.get('if', None) is None:
        return True

    return render_variable(context, wrap_jinja_braces(hook_dict['if']))


def merge_block_output(
    hook_output_value: Any,
    context: Context,
    append_hook_value: bool = False,
):
    """
    Block hooks have already written to the output dict so to merge, need to take the
     keys from the key path and move them up one level.
    """
    if append_hook_value:
        # TODO: https://github.com/robcxyz/tackle-box/issues/66
        #  Allow merging into lists
        if isinstance(context.key_path_block[-1], bytes):
            # An exception maybe needed here or this error is snubbed.
            pass
        raise exceptions.AppendMergeException(
            "Can't merge from for loop.", context=context
        ) from None

    # 66 - Should qualify dict here
    target_context, key_path = get_target_and_key(context)
    indexed_block_output = nested_get(element=hook_output_value, keys=key_path)
    for k, v in indexed_block_output.items():
        element, key_path = get_target_and_key(context)
        nested_set(
            element=target_context,
            keys=key_path[:-1] + [k],
            value=v,
        )
    nested_delete(element=target_context, keys=key_path)


def merge_output(
    hook_output_value: Any,
    context: Context,
    append_hook_value: bool = False,
):
    """Merge the contents into it's top level set of keys."""
    if append_hook_value:
        raise exceptions.AppendMergeException(
            "Can't merge from for loop.", context=context
        ) from None

    if context.key_path[-1] in ('->', '_>'):
        # Expanded key - Remove parent key from key path
        key_path = context.key_path[:-2] + [context.key_path[-1]]
    else:
        # Compact key
        key_path = context.key_path[:-1] + [context.key_path[-1][-2:]]

    # Can't merge into top level keys without merging k/v individually
    if len(key_path) == 1:
        # This is only valid for dict output
        if isinstance(hook_output_value, (dict, OrderedDict)):
            for k, v in hook_output_value.items():
                set_key(context=context, value=v, key_path=[k] + key_path)
            return
        else:
            raise exceptions.TopLevelMergeException(
                "Can't merge non maps into top level keys.", context=context
            ) from None
    else:
        if isinstance(hook_output_value, dict):
            for k, v in hook_output_value.items():
                set_key(context=context, value=v, key_path=key_path + [k])
            return
        else:
            set_key(context=context, value=hook_output_value, key_path=key_path)


def run_hook_in_dir(hook: Type[BaseHook]) -> Any:
    if hook.chdir:
        path = os.path.abspath(os.path.expanduser(hook.chdir))
        if os.path.isdir(path):
            # Use contextlib to switch dirs
            with work_in(os.path.abspath(os.path.expanduser(hook.chdir))):
                return hook.exec()
        else:
            raise exceptions.HookUnknownChdirException(
                f"The specified path='{path}' to change to was not found.",
                hook=hook,
            ) from None
    else:
        return hook.exec()


def render_hook_vars(hook_dict: dict, Hook: ModelMetaclass, context: 'Context'):
    """Render the hook variables."""
    for key, value in list(hook_dict.items()):
        if key not in Hook.__fields__ and key not in BASE_METHODS:
            # If the hook has a `kwargs` field, then map it to that field.
            if Hook.__fields__['kwargs'].default is not None:
                default_kwargs = Hook.__fields__['kwargs'].default
                if default_kwargs not in hook_dict:
                    hook_dict[default_kwargs] = {}
                hook_dict[default_kwargs][key] = value
                hook_dict.pop(key)
                continue

            # Get a list of possible fields for hook before raising error.
            possible_fields = [
                f"{k}: {v.type_.__name__}"
                for k, v in Hook.__fields__.items()
                if k not in BaseHook.__fields__
            ]
            # TODO: Include link to docs -> Will need to also include provider name
            #  and then differentiate between lazy imported hooks which already have the
            #  provider and the ones that don't
            raise exceptions.UnknownInputArgumentException(
                f"Key={key} not in hook={hook_dict['hook_type']}. Possible values are "
                f"{', '.join(possible_fields)}",
                context=context,
            ) from None

        if key in Hook.Config.alias_to_fields:
            # Skip any keys used in logic as these are evaluated / rendered separately
            continue

        if key in Hook._render_exclude or key in Hook._render_exclude_default:
            # Skip anything that has been marked excluded. Needed for things like block
            # hooks which will have templating within the inputs
            continue

        if isinstance(value, str):
            # Check Hook private vars for rendering by default which wraps bare strings
            if key in Hook._render_by_default:
                hook_dict[key] = render_variable(context, wrap_jinja_braces(value))

            # TODO: When we build our own custom Field function then this will change
            # TODO: This causes errors when the field is aliased as the lookup doesn't
            #  work and needs a deeper introspection.
            #  https://github.com/robcxyz/tackle-box/issues/80
            #  Fixing with custom Field def should fix this.
            elif 'render_by_default' in Hook.__fields__[key].field_info.extra:
                hook_dict[key] = render_variable(context, wrap_jinja_braces(value))

            elif ('{{' in value and '}}' in value) or ('{%' in value and '%}'):
                hook_dict[key] = render_variable(context, value)

        elif isinstance(value, (list, dict)):
            hook_dict[key] = render_variable(context, value)


def parse_sub_context(
    context: 'Context', hook_dict: dict, target: str, append_hook_value: bool = None
):
    """
    Reparse a subcontext as in the case with `else` and `except` where you have to
    handle the negative side of the either `if` or `try`. Works on both looped and
    normal runs by checking the last item in the key path. Then overwrites the input
    with a new context.
    """
    hook_target = hook_dict[target]
    if isinstance(hook_target, str):
        set_key(
            context=context,
            value=render_variable(context, hook_dict[target]),
            append_hook_value=append_hook_value,
        )
        return
    elif isinstance(hook_target, (bool, int, float)):
        set_key(
            context=context,
            value=hook_target,
            append_hook_value=append_hook_value,
        )
        return

    indexed_key_path = context.key_path[
        (len(context.key_path_block) - len(context.key_path)) :  # noqa
    ]

    if isinstance(indexed_key_path[-1], bytes):
        # We are in a for loop
        input_dict = nested_get(
            element=context.input_context,
            keys=indexed_key_path[:-3],
        )
        updated_item = [
            hook_dict[target] if i == decode_list_index(context.key_path[-1]) else None
            for i in range(decode_list_index(context.key_path[-1]) + 1)
        ]
        # TODO: Figure out wtf is going on here...
        input_dict[indexed_key_path[-3]] = updated_item
        walk_sync(
            context,
            element=input_dict[indexed_key_path[-3]][
                decode_list_index(context.key_path[-1])
            ],
        )

    else:
        input_dict = nested_get(
            element=context.input_context,
            keys=indexed_key_path[:-2],
        )
        arrow = context.key_path[-1]
        input_dict[indexed_key_path[-2]] = {arrow: 'block', 'items': hook_dict[target]}
        walk_sync(context, element=input_dict[indexed_key_path[-2]])


def parse_hook(
    hook_dict, Hook: ModelMetaclass, context: 'Context', append_hook_value: bool = None
):
    """Parse input dict for loop and when logic and calls hooks."""
    if evaluate_if(hook_dict, context, append_hook_value):

        if 'for' in hook_dict:
            # This runs the current function in a loop with `append_hook_value` set so
            # that keys are appended in the loop.
            evaluate_for(hook_dict, Hook, context)
            return

        else:
            # Render the remaining hook variables
            render_hook_vars(hook_dict, Hook, context)

            try:
                hook = Hook(
                    **hook_dict,
                    input_context=context.input_context,
                    public_context=context.public_context,
                    private_context=context.private_context,
                    temporary_context=context.temporary_context,
                    existing_context=context.existing_context,
                    no_input=context.no_input,
                    calling_directory=context.calling_directory,
                    calling_file=context.calling_file,
                    public_hooks=context.public_hooks,
                    private_hooks=context.private_hooks,
                    key_path=context.key_path,
                    verbose=context.verbose,
                    env_=context.env_,
                    is_hook_call=True,
                )
            except ValidationError as e:
                # Handle any try / except logic
                if 'try' in hook_dict and hook_dict['try']:
                    if 'except' in hook_dict and hook_dict['except']:
                        parse_sub_context(context, hook_dict, target='except')
                    return

                msg = str(e)
                if Hook.identifier.startswith('tackle.providers'):
                    id_list = Hook.identifier.split('.')
                    provider_doc_url_str = id_list[2].title()
                    # Replace the validated object name (ex PrintHook) with the
                    # hook_type field that users would more easily know.
                    msg = msg.replace(id_list[-1], f"{hook_dict['hook_type']} hook")

                    msg += (
                        f"\n Check the docs for more information on the hook -> "
                        f"https://robcxyz.github.io/tackle-box/providers/"
                        f"{provider_doc_url_str}/{hook_dict['hook_type']}/"
                    )
                raise exceptions.HookParseException(str(msg), context=context) from None

            # Main exec logic
            if hook.try_:
                try:
                    hook_output_value = run_hook_in_dir(hook)
                except Exception as e:
                    if hook.verbose:
                        print(e)
                    if hook.except_:
                        parse_sub_context(context, hook_dict, target='except')
                    return
            else:
                # Normal hook run
                hook_output_value = run_hook_in_dir(hook)

            if hook.skip_output:
                if hook.merge:
                    merge_block_output(
                        hook_output_value=hook_output_value,
                        context=context,
                        append_hook_value=append_hook_value,
                    )
                return
            elif hook.merge:
                merge_output(
                    hook_output_value=hook_output_value,
                    context=context,
                    append_hook_value=append_hook_value,
                )
            else:
                set_key(
                    context=context,
                    value=hook_output_value,
                    append_hook_value=append_hook_value,
                )

    elif 'else' in hook_dict:
        parse_sub_context(context, hook_dict, target='else')


def evaluate_args(
    args: list,
    hook_dict: dict,
    Hook: ModelMetaclass,  # noqa
    context: 'BaseContext' = None,  # For error handling
):
    """
    Associate hook arguments provided in the call with hook attributes. Parses the
    hook's `args` attribute to know how to map arguments are mapped to where and
    deal with rendering by default.

    TODO: This needs to be re-thought. Right now we parse the inputs without regard
     for the types of the argument mapping. What could be better is if we know the types
     of the arg mapping ahead of time and then try to assemble the most logical mapping
     afterwards. So if the mapping consists of a [str, list], then the if the first
     args are strs then we can ignore the list part. Right now it would just join all
     the strings together if they are part of last arg mapping.
    TODO: Improve error handling.
    Solutions:
    - First try to infer type from arg
        - Single types then into unions?
    - If type cannot be infered (ie Any) then do ast as literal
    """
    hook_args: list = Hook.__fields__['args'].default
    for i, v in enumerate(args):
        # Iterate over the input args
        if i + 1 == len(hook_args):
            # We are at the last argument mapping so we need to join the remaining
            # arguments as a single string if it is not a list of another map.
            if not isinstance(args[i], (str, float)):
                # Catch list dict and ints - strings floats and bytes caught later
                value = args[i]
            elif Hook.__fields__[hook_args[i]].type_ == str:
                # Was parsed on spaces so reconstructed.
                value = ' '.join(args[i:])
            elif Hook.__fields__[hook_args[i]].type_ in (bool, float, int):
                # TODO: Incomplete
                value = args[i]
            elif isinstance(Hook.__fields__[hook_args[i]], list):
                # If list then all the remaining items
                value = args[i:]
            elif isinstance(v, str):
                # Make assumption the rest of the args can be reconstructed as above
                value = ' '.join(args[i:])
            elif isinstance(v, (bool, float, int)):
                # TODO: Incomplete
                if len(args[i:]) > 1:
                    value = args[i:]
                else:
                    value = args[i]
            else:
                # Only thing left is a dict
                if len(args[i:]) > 1:
                    raise exceptions.HookParseException(
                        f"Can't specify multiple arguments for map argument "
                        f"{Hook.__fields__[hook_args[i]]}.",
                        context=context,
                    ) from None
                value = args[i]
            hook_dict[hook_args[i]] = value
            args.pop(0)
            return
        else:
            # The hooks arguments are indexed
            try:
                hook_dict[hook_args[i]] = v
            except IndexError:
                if len(hook_args) == 0:
                    # TODO: Give more info on possible methods
                    hook_name = Hook.identifier.split('.')[-1]
                    if hook_name == '':
                        hook_name = 'default'
                    raise exceptions.UnknownArgumentException(
                        f"The {hook_name} hook does not take any "
                        f"arguments. Hook argument {v} caused an error.",
                        context=context,
                    ) from None
                else:
                    raise exceptions.UnknownArgumentException(
                        f"The hook {hook_dict['hook_type']} takes the following indexed"
                        f"arguments -> {hook_args} which does not map to the arg {v}.",
                        context=context,
                    ) from None


def run_hook(context: 'Context'):
    """
    Run the hook by qualifying the input argument and matching the input params with the
     hook's `_args` which are then overlayed into a hook kwargs. Also interprets
     special cases where you have a string or list input of renderable variables.
    """
    if isinstance(context.input_string, str):
        args, kwargs, flags = unpack_args_kwargs_string(context.input_string)
        args = var_hook_macro(args)
        # Remove first args it will be consumed and no longer relevant
        first_arg = args.pop(0)

    else:
        # Rare case when an arrow is used to indicate rendering of a list.
        # Only qualified when input is of form `key->: [{{var}},{{var}},...]
        # In this case we need to set the key as an empty list
        set_key(context, value=[])

        # Iterate over values appending rendered values. Rendered values can be any type
        for i, v in enumerate(context.input_string):
            set_key(
                context=context,
                value=render_variable(context, v),
                key_path=context.key_path + [encode_list_index(i)],
            )
        return

    # Look up the hook from the imported providers
    Hook = get_hook(first_arg, context)

    if context.key_path[-1] in ('->', '_>'):
        # We have an expanded or mixed (with args) hook expression and so there will be
        # additional properties in adjacent keys. Trim key_path_block for blocks
        try:
            hook_dict = nested_get(
                context.input_context,
                smush_key_path(context.key_path[:-1][len(context.key_path_block) :]),
            ).copy()
        except KeyError as e:
            raise exceptions.UnknownHookTypeException(
                f"Key: {e} - Unknown", context=context
            ) from None
        # Need to replace arrow keys as for the time being (pydantic 1.8.2) - multiple
        # aliases for the same field (type) can't be specified so doing this hack
        if '->' in hook_dict:
            hook_dict.pop('->')
        else:
            hook_dict.pop('_>')
    else:
        # Hook is a compact expression - Can only be a string
        hook_dict = {}
    hook_dict['hook_type'] = first_arg

    if 'args' in kwargs:
        # For calling hooks, you can manually provide the hook with args. Useful for
        # creating declarative hooks that
        hook_args = kwargs.pop('args')
        if isinstance(hook_args, list):
            args += hook_args
        else:
            args += [hook_args]

    # Associate hook arguments provided in the call with hook attributes
    evaluate_args(args=args, hook_dict=hook_dict, Hook=Hook, context=context)
    # Add any kwargs
    for k, v in kwargs.items():
        hook_dict[k] = v
    for i in flags:
        hook_dict[i] = True

    if 'kwargs' in hook_dict:
        hook_kwargs = hook_dict.pop('kwargs')
        if isinstance(hook_kwargs, dict):
            hook_dict.update(hook_kwargs)
        elif isinstance(hook_kwargs, str):
            try:
                hook_dict.update(
                    render_variable(context=context, raw=wrap_jinja_braces(hook_kwargs))
                )
            except ValueError:
                error_msg = (
                    "The parameter `kwargs` should be either a map or a string "
                    "reference to a map."
                )
                raise exceptions.UnknownArgumentException(
                    error_msg, context=context
                ) from None
        else:
            error_msg = (
                "The parameter `kwargs` should be either a map or a string reference "
                "to a map."
            )
            raise exceptions.UnknownArgumentException(
                error_msg, context=context
            ) from None

    # Cleanup any unquoted fields -> common mistake that is hard to debug producing a
    #  nested dict that breaks parsing / hook calls. Ex foo: {{bar}} -> foo: "{{bar}}"
    cleanup_unquoted_strings(hook_dict)

    # Main parser
    parse_hook(hook_dict, Hook, context)


def walk_sync(context: 'Context', element):
    """
    Traverse an object looking for hook calls and running those hooks. Here we are
     keeping track of which keys are traversed in a list called `key_path` with strings
     as dict keys and byte encoded integers for list indexes.
    """
    if len(context.key_path) != 0:
        # Handle compact expressions - ie key->: hook_type args
        ending = context.key_path[-1][-2:]
        if ending in ('->', '_>'):
            if isinstance(element, str):
                element = compact_hook_call_macro(context, element)
            elif isinstance(element, list):
                element = list_to_var_macro(context, element)

    if isinstance(element, dict):
        # Handle expanded expressions - ie key:\n  ->: hook_type args
        if '->' in element.keys():
            # Public hook calls
            context.key_path.append('->')
            context.input_string = element['->']
            run_hook(context)
            context.key_path.pop()
            return
        elif '_>' in element.keys():
            # Private hook calls
            context.key_path.append('_>')
            context.input_string = element['_>']
            run_hook(context)
            context.key_path.pop()
            return
        elif element == {}:
            set_key(context=context, value={})
            return

        for k, v in element.copy().items():

            if isinstance(v, CommentedMap):
                # This is for a common parsing error that messes up values with braces.
                # For instance `stuff->: {{things}}` (no quotes), ruamel interprets as
                # 'stuff': ordereddict([(ordereddict([('things', None)]), None)]) which
                # technically is accurate but generally users would never actually do.
                # Since it is common to forget to quote, this is a helper to try to
                # catch that error and fix it.  Warning -> super hacky....
                if len(v) == 1:
                    value_ = next(iter(v.values()))
                    key_ = next(iter(v.keys()))
                    if value_ is None and isinstance(key_, CommentedKeyMap):
                        if context.verbose:
                            _key_path = get_readable_key_path(context.key_path)
                            msg = f"Handling unquoted template at key path {_key_path}."
                            print(msg)
                        v = "{{" + next(iter(next(iter(v.keys())))) + "}}"

            context.key_path.append(k)
            # Special case where we have an empty hook, expanded or compact
            if k[-2:] in ('->', '_>') and (v is None or isinstance(v, dict)):
                # Here we re-write the input to turn empty hooks into block hooks
                blocks_macro(context)
                context.key_path[-1] = k[:-2]
                value = nested_get(
                    element=context.input_context,
                    keys=context.key_path[
                        (len(context.key_path_block) - len(context.key_path)) :
                    ],
                )
                # Finally check if the `items` key exists in the input_context.  If not
                #  then we have an empty hook which will cause an ambiguous
                #  ValidationError for missing field
                if 'items' not in value:
                    raise exceptions.EmptyBlockException(
                        "Empty block hook.", context=context
                    ) from None

                walk_sync(context, value)
                context.key_path.pop()
            else:
                # Recurse
                walk_sync(context, v)
                context.key_path.pop()
    # Non-hook calls recurse through inputs
    elif isinstance(element, list):
        # Handle empty lists
        if len(element) == 0:
            set_key(context=context, value=element)
        else:
            for i, v in enumerate(element.copy()):
                context.key_path.append(encode_list_index(i))
                walk_sync(context, v)
                context.key_path.pop()
    else:
        set_key(context=context, value=element)


def update_input_context_with_kwargs(context: 'Context', kwargs: dict):
    """
    Update the input dict with kwargs which in this context are treated as overriding
    the keys. Takes into account if the key is a hook and replaces that.
    """
    for k, v in kwargs.items():
        if k in context.input_context:
            context.input_context.update({k: v})
        elif f"{k}->" in context.input_context:
            # Replace the keys and value in the same position it was in
            context.input_context = {
                key if key != f"{k}->" else k: value if key != f"{k}->" else v
                for key, value in context.input_context.items()
            }


def update_hook_with_kwargs_and_flags(hook: ModelMetaclass, kwargs: dict) -> dict:
    """
    For consuming kwargs / flags, once the hook has been identified when calling hooks
    via CLI actions, this function matches the kwargs / flags with the hook and returns
    any unused kwargs / flags for use in the outer context. Note that flags are kwargs
    as they have already been merged by now.
    """
    for k, v in kwargs.copy().items():
        if k in hook.__fields__:
            if hook.__fields__[k].type_ == bool:
                # Flags -> These are evaluated as the inverse of whatever is the default
                if hook.__fields__[k].default:  # ie -> True
                    hook.__fields__[k].default = False
                else:
                    hook.__fields__[k].default = True
            else:
                # Kwargs
                hook.__fields__[k].default = v
            hook.__fields__[k].required = False  # Otherwise will complain
            kwargs.pop(k)
    return kwargs


def find_run_hook_method(
    context: 'Context', hook: ModelMetaclass, args: list, kwargs: dict
) -> Any:
    """
    Given a hook with args, find if the hook has methods and if it does not, apply the
     args to the hook based on the `args` field mapping. Calls the hook.
    """
    kwargs = update_hook_with_kwargs_and_flags(
        hook=hook,
        kwargs=kwargs,
    )

    if kwargs != {}:
        # We were given extra kwargs / flags so should throw error
        hook_name = hook.identifier.split('.')[-1]
        if hook_name == '':
            hook_name = 'default'
        unknown_args = ' '.join([f"{k}={v}" for k, v in kwargs.items()])
        raise exceptions.UnknownInputArgumentException(
            f"The args {unknown_args} not recognized when running the hook/method "
            f"{hook_name}. Exiting.",
            context=context,
        ) from None

    arg_dict = {}
    num_popped = 0
    for i, arg in enumerate(args.copy()):
        if arg in hook.__fields__ and hook.__fields__[arg].type_ == Callable:
            # Consume the args
            args.pop(i - num_popped)
            num_popped += 1

            # Gather the function's dict so it can be compiled into a runnable hook
            func_dict = hook.__fields__[arg].default.function_dict.copy()

            # Add inheritance from base function fields
            for j in hook.__fields__['function_fields'].default:
                # Base method should not override child.
                if j not in func_dict:
                    func_dict[j] = hook.__fields__[j]

            new_hook = create_function_model(
                context=context,
                func_name=hook.__fields__[arg].name,
                func_dict=func_dict,
            )

            hook = new_hook

        elif arg == 'help':
            # Exit 0
            run_help(context=context, hook=hook)
        elif 'args' in hook.__fields__:
            evaluate_args(args, arg_dict, Hook=hook, context=context)
        else:
            raise exceptions.UnknownInputArgumentException(
                "Can't find the ", context=context
            ) from None

    return hook(**kwargs, **arg_dict).exec()


def raise_if_args_exist(
    context: 'Context', hook: ModelMetaclass, args: list, kwargs: dict, flags: list
):
    """
    Raise an error if not all the args / kwargs / flags have been consumed which would
     mean the user supplied extra vars and should be yelled at.
    """
    msgs = []
    if len(args) != 0:
        msgs.append(f"args {', '.join(args)}")
    if len(kwargs) != 0:
        missing_kwargs = ', '.join([f"{k}={v}" for k, v in kwargs.items()])
        msgs.append(f"kwargs {missing_kwargs}")
    if len(flags) != 0:
        msgs.append(f"flags {', '.join(flags)}")
    if len(msgs) != 0:
        if hook:
            hook_name = hook.identifier.split('.')[-1]
            if hook_name == '':
                hook_name = 'default'
            raise exceptions.UnknownInputArgumentException(
                # TODO: Add the available args/kwargs/flags to this error msg
                f"The {' and '.join(msgs)} were not found in the \"{hook_name}\" hook. "
                f"Run the same command without the arg/kwarg/flag + \"help\" to see the "
                f"available args/kwargs/flags.",
                context=context,
            ) from None
        else:
            raise exceptions.UnknownSourceException(
                f"Could not find source = {args[0]} or as key / hook in parent tackle "
                f"file.",
                context=context,
            ) from None


def run_source(context: 'Context', args: list, kwargs: dict, flags: list) -> Optional:
    """
    Process global args/kwargs/flags based on if the args relate to the default hook or
     some public hook (usually declarative). Once the hook has been identified, the
     args/kwargs/flags are consumed and if there are any args left, an error is raised.
    """
    # Tackle is called both through the CLI and as a package and so to preserve args /
    # kwargs we merge them here.
    if context.global_args is not None:
        args = args + context.global_args
        context.global_args = None

    # Global kwargs/flags are immediately consumed and injected into the kwargs/flags
    if context.global_kwargs is not None:
        kwargs.update(context.global_kwargs)
        context.global_kwargs = None

    if context.global_flags is not None:
        # TODO: Validate -> These are temporarily set to true but will be resolved as
        #  the inverse of what the default is
        kwargs.update({i: True for i in context.global_flags})
        context.global_flags = None

    # For CLI calls, this logic lines up the args with methods / method args and
    # integrates the kwargs / flags into the call
    if len(args) == 0 and context.default_hook:  # Default hook (no args)
        # Add kwargs / flags (already merged into kwargs) to default hook
        kwargs = update_hook_with_kwargs_and_flags(
            hook=context.default_hook,
            kwargs=kwargs,
        )
        # Run the default hook as there are no args. The outer context is then parsed
        #  as otherwise it would be unreachable.
        try:
            default_hook_output = context.default_hook().exec()
        except ValidationError as e:
            raise exceptions.UnknownInputArgumentException(
                e.__str__(), context=context
            ) from None
        # Return the output of the default hook
        # TODO: Determine what the meaning of a primitive type return means with some
        #  kind of outer context -> Should error (and be caught) if there is a conflict
        #  in types.
        context.public_context = default_hook_output
        # TODO: ??? -> If output is primitive, then we need to return it without parsing
        #  the outer context
        raise_if_args_exist(  # Raise if there are any args left
            context=context,
            hook=context.default_hook,
            args=args,
            kwargs=kwargs,
            flags=flags,
        )
    elif len(args) == 1 and args[0] == 'help' and context.default_hook is not None:
        run_help(context=context, hook=context.default_hook)
    elif len(args) == 1 and args[0] == 'help':
        run_help(context=context)
    elif len(args) != 0:  # With args
        # Prioritize public_hooks (ie non-default hook) because if the hook exists,
        # then we should consume the arg there instead of using the arg as an arg for
        # default hook because otherwise the public hook would be unreachable.
        if args[0] in context.public_hooks:  #
            # Search within the public hook for additional args that could be
            # interpreted as methods which always get priority over consuming the arg
            # as an arg within the hook itself.
            public_hook = args.pop(0)  # Consume arg
            context.public_context = find_run_hook_method(
                context=context,
                hook=context.public_hooks[public_hook],
                args=args,
                kwargs=kwargs,
            )
            raise_if_args_exist(  # Raise if there are any args left
                context=context,
                hook=context.public_hooks[public_hook],
                args=args,
                kwargs=kwargs,
                flags=flags,
            )
        elif context.default_hook:
            context.public_context = find_run_hook_method(
                context=context, hook=context.default_hook, args=args, kwargs=kwargs
            )
            raise_if_args_exist(  # Raise if there are any args left
                context=context,
                hook=context.default_hook,
                args=args,
                kwargs=kwargs,
                flags=flags,
            )
        else:
            raise_if_args_exist(  # Raise if there are any args left
                context=context,
                hook=context.default_hook,  # This is None if it does not exist
                args=args,
                kwargs=kwargs,
                flags=flags,
            )
        # We are not going to parse the outer context in this case as we are already
        # within a hook. Doesn't necessarily make sense in this case.
        return
    else:
        # If there are no declarative hooks defined, use the kwargs to override values
        #  within the context.
        update_input_context_with_kwargs(context=context, kwargs=kwargs)

    for i in flags:
        # Process flags by setting key to true
        context.input_context.update({i: True})

    if len(context.input_context) == 0:
        if context.public_context is None:
            raise exceptions.EmptyTackleFileException(
                # TODO improve -> Should give help by default?
                f"Only functions are declared in {context.input_string} tackle file. Must"
                f" provide an argument such as [] or run `tackle {context.input_string}"
                f" help` to see more options.",
                context=context,
            ) from None
    else:
        walk_sync(context, context.input_context.copy())


def parse_tmp_context(context: Context, element: Any, existing_context: dict):
    """
    Parse an arbitrary element. Only used for declarative hook field defaults and in
     the `run_hook` hook in the tackle provider.
    """
    tmp_context = Context(
        public_hooks=context.public_hooks,
        private_hooks=context.private_hooks,
        public_context={},
        private_context={},
        temporary_context=context.temporary_context,
        existing_context=existing_context,
        input_context=element,
        key_path=['->'],
        key_path_block=['->'],
        no_input=context.no_input,
        calling_directory=context.calling_directory,
        calling_file=context.calling_file,
        verbose=context.verbose,
    )
    walk_sync(context=tmp_context, element=element)

    return tmp_context.public_context


def function_walk(
    self: Type[ModelMetaclass],
    input_element: Union[list, dict],
    return_: Union[list, dict] = None,
) -> Any:
    """
    Walk an input_element for a function and either return the whole context or one or
     many returnable string keys. Function is meant to be implanted into a function
     object and called either as `exec` or some other arbitrary method.
    """
    if input_element is None:
        # If there is no `exec` method, input_element is None so we infer that the
        # input fields are to be returned. This is useful if the user would like to
        # validate a dict easily with a function and is the only natural meaning of
        # a function call without an exec method.
        input_element = {}
        for i in self.function_fields:
            input_element[i] = getattr(self, i)

    if self.public_context:
        existing_context = self.public_context.copy()
        existing_context.update(self.existing_context)
    else:
        existing_context = {}

    for i in self.function_fields:
        # Update a function's existing context with the already matched args
        value = getattr(self, i)
        if isinstance(value, dict) and '->' in value:
            # For when the default has a hook in it
            output = parse_tmp_context(
                context=self, element={i: value}, existing_context=existing_context
            )
            existing_context.update(output)
            input_element[i] = output[i]
        else:
            existing_context.update({i: getattr(self, i)})

    tmp_context = Context(
        public_hooks=self.public_hooks,
        private_hooks=self.private_hooks,
        existing_context=existing_context,
        public_context={},
        input_context=input_element,
        key_path=[],
        no_input=self.no_input,
        calling_directory=self.calling_directory,
        calling_file=self.calling_file,
    )
    walk_sync(context=tmp_context, element=input_element.copy())

    if return_:
        if isinstance(return_, str):
            if return_ in tmp_context.public_context:
                return tmp_context.public_context[return_]
            else:
                raise exceptions.FunctionCallException(
                    f"Return value '{return_}' is not found " f"in output.",
                    function=self,
                ) from None
        elif isinstance(return_, list):
            if isinstance(tmp_context, list):
                # TODO: This is not implemented (ie list outputs)
                raise exceptions.FunctionCallException(
                    f"Can't have list return {return_} for " f"list output.",
                    function=self,
                ) from None
            output = {}
            for i in return_:
                # Can only return top level keys right now
                if i in tmp_context.public_context:
                    output[i] = tmp_context.public_context[i]
                else:
                    raise exceptions.FunctionCallException(
                        f"Return value '{i}' in return {return_} not found in output.",
                        function=self,
                    ) from None
            return tmp_context.public_context[return_]
        else:
            raise NotImplementedError(f"Return must be of list or string {return_}.")
    return tmp_context.public_context


def create_function_model(
    context: 'Context', func_name: str, func_dict: dict
) -> Type[BaseFunction]:
    """Create a model from the function input dict."""
    if func_dict is None:
        raise exceptions.EmptyFunctionException(
            "Can't have an empty function", context=context, function_name=func_name
        )
    # Macro to expand all keys properly so that a field's default can be parsed
    func_dict = function_field_to_parseable_macro(func_dict, context, func_name)

    if func_name.endswith(('<-', '<_')):
        func_name = func_name[:-2]

    # Implement inheritance
    if 'extends' in func_dict and func_dict['extends'] is not None:
        base_hook = get_hook(func_dict['extends'], context)
        func_dict = {**base_hook().function_dict, **func_dict}
        func_dict.pop('extends')

    # Persisted with object for `extends`. Used later
    function_dict = func_dict.copy()

    # fmt: off
    # Validate raw input params against pydantic object where values will be used later
    exec_ = None
    if 'exec' in func_dict:
        exec_ = func_dict.pop('exec')
    elif 'exec<-' in func_dict:
        exec_ = func_dict.pop('exec<-')

    # Special vars
    function_input = FunctionInput(
        exec_=exec_,
        return_=func_dict.pop('return') if 'return' in func_dict else None,
        args=func_dict.pop('args') if 'args' in func_dict else [],
        render_exclude=func_dict.pop(
            'render_exclude') if 'render_exclude' in func_dict else [],
        help=func_dict.pop('help') if 'help' in func_dict else None,
        # TODO: Build validators
        # validators_=func_dict.pop('validators') if 'validators' in func_dict else None,
    )

    # fmt: on
    new_func = {'hook_type': func_name, 'function_fields': []}
    literals = ('str', 'int', 'float', 'bool', 'dict', 'list')  # strings to match
    # Create function fields from anything left over in the function dict
    for k, v in func_dict.items():
        if v is None:
            # TODO: Why skip? Would be ignored. Empty keys mean something right?
            continue

        if k.endswith(('->', '_>')):
            raise NotImplementedError
        elif k.endswith(('<-', '<_')):
            # Implement method which is instantiated later in `get_hook`
            new_func[k[:-2]] = (Callable, LazyBaseFunction(function_dict=v))
            continue

        elif isinstance(v, dict):
            if 'type' in v:
                # TODO: Qualify type in enum -> Type
                type_ = v['type']
                if type_ not in literals:
                    raise exceptions.MalformedFunctionFieldException(
                        f"Function field {k} with type={v} unknown. Must be one of {','.join(literals)}",
                        function_name=func_name,
                        context=context,
                    ) from None
                if 'description' in v:
                    v = dict(v)
                    v['description'] = v['description'].__repr__()
                new_func[k] = (type_, Field(**v))
            elif 'default' in v:
                if isinstance(v['default'], dict) and '->' in v['default']:
                    # For hooks in the default fields.
                    new_func[k] = (Any, Field(**v))
                else:
                    new_func[k] = (type(v['default']), Field(**v))
            else:
                raise exceptions.MalformedFunctionFieldException(
                    f"Function field {k} must have either a `type` or `default` field "
                    f"where the type can be inferred.",
                    function_name=func_name,
                    context=context,
                ) from None
        elif v in literals:
            new_func[k] = (locate(v).__name__, Field(...))
        elif isinstance(v, (str, int, float, bool)):
            new_func[k] = v
        elif isinstance(v, list):
            new_func[k] = (list, v)
        elif isinstance(v, ModelField):
            # Is encountered when inheritance is imposed and calling function methods
            new_func[k] = (v.type_, v.default)
        else:
            raise Exception("This should never happen")

        # function_fields used later to populate functions without an exec method and
        # the context for rendering inputs.
        new_func['function_fields'].append(k)

    # TODO: Update module namespace
    # Create a function with the __module__ default to pydantic.main
    try:
        Function = create_model(
            func_name,
            __base__=BaseFunction,
            **new_func,
            **function_input.dict(include={'args', 'render_exclude'}),
            **{'function_dict': (dict, function_dict)},  # Preserve for `extends` key
            # https://github.com/robcxyz/tackle-box/issues/99
            **{'public_hooks': context.public_hooks},
            **{'private_hooks': context.private_hooks},
        )
    except NameError as e:
        if 'shadows a BaseModel attribute' in e.args[0]:
            shadowed_arg = e.args[0].split('\"')[1]
            extra = "a different value"
            raise exceptions.ShadowedFunctionFieldException(
                f"The function field \'{shadowed_arg}\' is reserved. Use {extra}.",
                function_name=func_name,
                context=context,
            ) from None
        # Don't know what else could happen
        raise e

    # Create an 'exec' method on the function that can be called later.
    setattr(
        Function,
        'exec',
        partialmethod(function_walk, function_input.exec_, function_input.return_),
    )

    # TODO: Rm when filters fixed
    # context.env_.filters[func_name] = Function(
    #     existing_context={},
    #     no_input=context.no_input,
    # ).wrapped_exec

    return Function


def extract_functions(context: 'Context'):
    """
    Iterate through all the keys along baseline of tackle file and extract / compile
     all the keys that reference functions.
    """
    for k, v in context.input_context.copy().items():
        if re.match(r'^[a-zA-Z0-9\_]*(<\-|<\_)$', k):
            # TODO: RM arrow and put in associated access modifier namespace
            Function = create_function_model(context, k, v)
            function_name = k[:-2]
            arrow = k[-2:]
            if function_name == "":
                # Function is the default hook
                context.default_hook = Function
                context.input_context.pop(k)
            elif arrow == '<-':  # public hook
                context.public_hooks[function_name] = Function
                context.input_context.pop(k)
            elif arrow == '<_':  # private hook
                context.private_hooks[function_name] = Function
                context.input_context.pop(k)
            else:
                raise  # Should never happen


def extract_base_file(context: 'Context'):
    """Read the tackle file and initialize input_context."""
    if context.find_in_parent:
        try:
            path = find_in_parent(context.input_dir, [context.input_file])
        except NotADirectoryError:
            raise exceptions.UnknownSourceException(
                f"Could not find {context.input_file} in {context.input_dir} or in "
                f"any of the parent directories.",
                context=context,
            ) from None
        context.input_file = os.path.basename(path)
        context.input_dir = os.path.dirname(path)
    else:
        path = os.path.join(context.input_dir, context.input_file)

    # Preserve the calling file which should be carried over from tackle calls
    if context.calling_file is None:
        context.calling_file = context.input_file
    context.current_file = path

    try:
        context.input_context = read_config_file(path)
    except FileNotFoundError:
        raise exceptions.UnknownSourceException(
            f"Could not find file in {path}.", context=context
        ) from None
    except ParserError as e:
        raise exceptions.TackleFileInitialParsingException(e) from None

    if context.input_context is None:
        raise exceptions.EmptyTackleFileException(
            f"Tackle file found at {path} is empty.", context=context
        ) from None

    # Import the hooks and install requirements.txt if there is a ModuleNotFound error
    import_from_path(context, context.input_dir, skip_on_error=False)

    if isinstance(context.input_context, list):
        # Change output to empty list
        context.public_context = []
        context.private_context = []
    else:
        extract_functions(context)
        context.public_context = {}
        context.private_context = {}


def import_local_provider_source(context: 'Context', provider_dir: str):
    """
    Import a provider from a path by checking if the provider has a tackle file and
    returning a path.
    """
    context.input_dir = provider_dir
    if context.input_file is None:
        context.input_file = find_tackle_file(provider_dir)

    if context.directory:
        context.input_file = os.path.join(context.input_file, context.directory)

    extract_base_file(context)


def update_source(context: 'Context'):
    """
    Locate the repository directory from a template reference. This is the main parser
    for determining the source of the context and calls the succeeding parsing
    functions. The parsing order has the following order of precedence.

    If the template wasn't given then use the file in that parent directory.
    If the template refers to a zip file or zip url, download / unzip as the context.
    If the template refers to a repository URL, clone it.
    If the template refers to a file, use that as the context.
    """
    args, kwargs, flags = unpack_args_kwargs_string(context.input_string)

    first_arg = args[0]
    # Remove first args it will be consumed and no longer relevant
    args.pop(0)

    # Zipfile
    if first_arg.lower().endswith('.zip'):
        unzipped_dir = unzip(
            zip_uri=first_arg,
            clone_to_dir=settings.tackle_dir,
            no_input=context.no_input,
            password=context.password,  # TODO: RM - Should prompt?
        )
        import_local_provider_source(context, unzipped_dir)
    # Repo
    elif is_repo_url(first_arg):
        provider_dir = get_repo_source(first_arg, context.checkout)
        import_local_provider_source(context, provider_dir)
    # Directory
    elif is_directory_with_tackle(first_arg):
        # Special case where the input is a path to a directory. Need to override some
        # settings that would normally get populated by zip / repo refs. Does not need
        # a file reference as otherwise would be given absolute path to tackle file.
        context.input_file = os.path.basename(find_tackle_file(first_arg))
        context.input_dir = Path(first_arg).absolute()

        # Load the base file into input_context
        extract_base_file(context)
    # File
    elif is_file(first_arg):
        context.input_file = os.path.basename(first_arg)
        context.input_dir = Path(first_arg).parent.absolute()
        extract_base_file(context)
    # Search in parent
    else:
        # Lastly we check if there is a key in the parent that matches the arg. This is
        # basically the fallback logic but need to raise error right away if the key
        # does not exist so we don't have to catch it with context later.
        tackle_file = find_nearest_tackle_file()
        if tackle_file is None:
            raise exceptions.UnknownSourceException(
                f"Could not find source = {first_arg}", context=context
            ) from None

        context.input_file = os.path.basename(tackle_file)
        context.input_dir = Path(tackle_file).parent.absolute()
        extract_base_file(context)

        args.insert(0, first_arg)

    # We always change directory into the source that is being called. Needs to be this
    # or would be very confusing if writing a provider to always refer to it's own path.
    with work_in(context.input_dir):
        # Main parsing logic
        return run_source(context, args, kwargs, flags)



import os
from typing import Any
from pydantic import BaseModel, Field

from tackle.models import BaseHook
from tackle.hooks import import_from_path
from tackle.utils.vcs import get_repo_source


class RepoSource(BaseModel):
    """Repo object."""

    src: str
    version: str = None


class ImportHook(BaseHook):
    """
    Hook for importing external tackle-box providers. Does not actually execute the
     base tackle in the provider but merely makes the hooks and functions available to
     be used in the context. Takes any type as an argument to build `src` and
     `version` import targets.
    """

    hook_type: str = 'import'
    src: Any = Field(..., description="A str/list/dict as above.")
    version: str = Field(None, description="Version of src for remote imports.")
    args: list = ['src']

    def exec(self) -> None:
        if isinstance(self.src, str):
            # Get the provider path either local or remote.
            provider_path = self.get_dir_or_repo(self.src, self.version)
            import_from_path(self, provider_path)

        elif isinstance(self.src, list):
            for i in self.src:
                if isinstance(i, str):
                    import_from_path(self, i)
                if isinstance(i, dict):
                    # dict types validated above and transposed through same logic
                    repo_source = RepoSource(**i)

                    import_from_path(
                        self,
                        provider_path=self.get_dir_or_repo(
                            repo_source.src, repo_source.version
                        ),
                    )

        elif isinstance(self.src, dict):
            # Don't even check if path is directory as one would never use a dict here
            repo_source = RepoSource(**self.src)
            provider_dir = get_repo_source(
                repo=repo_source.src, repo_version=repo_source.version
            )
            self.provider_hooks.import_from_path(provider_path=provider_dir)

    def get_dir_or_repo(self, src, version):
        """Check if there is a path that matches input otherwise try as repo."""
        if os.path.exists(src):
            return src
        else:
            return get_repo_source(src, version)



from operator import itemgetter
from typing import Union, Optional

from tackle import BaseHook, Field
from tackle.utils.dicts import encode_key_path, nested_get
from tackle.exceptions import HookCallException


class SortHook(BaseHook):
    """
    Hook for sorting complex lists, dicts, or items within a key both in place or as
     output.
    """

    hook_type: str = 'sort'
    # fmt: off
    src: Union[list, dict, str] = Field(
        ...,
        description="Either a list of strings or a dict with keys to sort and return "
                    "the output or a string key_path to sort both in place or as "
                    "output (see `in_place`).")
    key: str = Field(
        None,
        description="If the `src` is a list of maps, the key to sort the contents by."
    )
    in_place: bool = Field(
        True, description="If the `src` is a string (ie a key path), then sort the "
                          "item in place (ie replace original) and return None.")
    reverse: bool = Field(False, description="To sort in reverse.")

    # TODO: Allow sorting the context based on key path
    # key: str = Field(...)
    src_is_key_path: bool = Field(
        False, description="If the src is a list and is meant to be a key path.")
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # fmt: on

    index: Union[int, list] = Field(
        None,
        description="If the input `src` is a list, use the index as the sort key. "
        "Takes both an int for single index or list for multiple criteria.",
    )
    args: list = ['src']

    def sort(self, src):
        if isinstance(src, list):
            if all(isinstance(n, dict) for n in src):
                # Iterrand is a dict
                return sorted(src, key=lambda d: d[self.key])

            if self.index is None:
                return src.sort(reverse=self.reverse)
            else:
                # TODO: Untested
                if isinstance(self.index, int):
                    # Source: https://stackoverflow.com/a/15544861/12642712
                    sorted_list = sorted(
                        src,
                        key=itemgetter(self.index),
                        reverse=self.reverse,
                    )
                    return sorted_list
                else:
                    # TODO: Untested
                    # TODO: wtf?
                    sorted_list = sorted(
                        src,
                        key=itemgetter(*self.index),
                        reverse=self.reverse,
                    )
                    return sorted_list
        else:
            # TODO: Untested
            # https://stackoverflow.com/a/613218/12642712
            sorted_dict = {
                k: v
                for k, v in sorted(
                    src.items(),
                    key=lambda item: item[1],
                    reverse=self.reverse,
                )
            }
            return sorted_dict

    def get_src(self):
        value = None
        contexts = ['public', 'private', 'temporary', 'existing']
        for i in contexts:
            try:
                value = nested_get(
                    element=getattr(self, f'{i}_context'),
                    keys=encode_key_path(self.src, self.sep),
                )
            except KeyError:
                pass

            if value is not None:
                break

        if value is None:
            raise HookCallException(
                f"Could not find {self.src} in any of the contexts (ie in "
                f"{','.join(contexts)}).",
                hook=self,
            )
        return value

    def exec(self) -> Optional[Union[list, dict]]:
        if (
            isinstance(self.src, str)
            or self.src_is_key_path
            and isinstance(self.src, list)
        ):
            self.src_is_key_path = True

            if self.in_place:
                self.src = self.get_src()
                return self.sort(self.src)
                # return self.src
            else:
                src = self.get_src()
                self.sort(src)
                return src
        else:
            # TODO: Untested
            self.sort(self.src)
            return self.src



"""Tests dict input objects for `tackle.providers.system.hooks.path` module."""
import os
from tackle.main import tackle


def test_provider_system_hook_path(change_dir):
    context = tackle('parent.yaml')

    assert context['path_isdir']

    assert context['path_exists']
    assert os.path.exists(context['find_in_parent_dir'])  # Should be tests/tests...
    assert os.path.exists(context['find_in_parent_file'])
    assert context['find_in_parent_fallback'] == context['find_in_parent_dir']


def test_provider_system_hook_path_child(change_dir):
    context = tackle('child.yaml')

    assert len(context['find_in_child']) == 2
    assert len(context['find_in_child_starting_dir']) == 1
    assert context['find_in_child_fallback'] == 'fallback.yaml'
    assert context['isfile']


def test_provider_paths_base_dir_name(change_dir):
    output = tackle('base-dir-name.yaml')

    assert output['base'] == 'tests'
    assert 'paths' in output['dir']



"""
Utils for modifying complex dictionaries generally based on an encoded key_path which is
a list of strings for key value lookups and byte encoded integers for items in a list.
"""
from typing import Union, Any, TYPE_CHECKING
from ruamel.yaml.constructor import CommentedKeyMap

if TYPE_CHECKING:
    from tackle.models import Context


def encode_list_index(list_index: int) -> bytes:
    """Encode an index for lists in a key path to bytes."""
    return list_index.to_bytes(2, byteorder='big')


def decode_list_index(list_index: bytes) -> int:
    """Decode an index for lists in a key path from bytes."""
    return int.from_bytes(list_index, byteorder='big')


def encode_key_path(path: Union[list, str], sep: str = "/") -> list:
    """
    Take a list or string key_path and encode it with bytes for items in a list. Strings
     are split up based on a separator, ie path/to/key.
    """
    if isinstance(path, str):
        path = path.split(sep)
        for i, v in enumerate(path):
            try:
                # Try to convert strings to ints
                path[i] = int(v)
            except ValueError:
                pass
        path = [i if not isinstance(i, int) else encode_list_index(i) for i in path]

    # Need to encode keys into bytes as that is how internally the parser works so
    # it doesn't jam up on any integer key. This hook will fail if the key is an
    # int.
    path = [i if not isinstance(i, int) else encode_list_index(i) for i in path]
    if isinstance(path, dict):
        # Could have key path expressed as dict?
        raise NotImplementedError

    return path


def get_key_from_key_path(key_path: list) -> str:
    """Take key_path and return the nearest key from end removing arrows."""
    if key_path[-1] in ('->', '_>'):
        # Expanded key
        return key_path[-2]
    elif key_path[-1].endswith('->', '_>'):
        # Compact key
        return key_path[-1][:-2]


def get_readable_key_path(key_path: list) -> str:
    """Take key_path and return the nearest key from end removing arrows."""
    readable_key_path = []
    for i in key_path:
        if i in ('->', '_>'):
            continue
            # return '.'.join(readable_key_path)
        elif i[-2:] in ('->', '_>'):
            readable_key_path.append(i[:-2])
            continue
        if isinstance(i, bytes):
            readable_key_path.append(str(decode_list_index(i)))
        else:
            readable_key_path.append(i)
    return '.'.join(readable_key_path)


def nested_delete(element, keys):
    """
    Delete items in a generic element (list / dict) based on a key path in the form of
    a list with strings for keys and byte encoded integers for indexes in a list.
    """
    num_elements = len(keys)

    if num_elements == 1:
        if isinstance(keys[0], bytes):
            element.pop(decode_list_index(keys[0]))
            return element

        elif isinstance(element, dict):
            element.pop(keys[0])
            return

    elif num_elements == 2:
        if isinstance(keys[0], bytes) and isinstance(keys[1], str):
            # Case for when we have an embedded item in a list but don't know if it is
            # a map with multiple keys that needs to have a single key removed or the
            # whole item itself
            # TODO
            if isinstance(element[decode_list_index(keys[0])], dict):
                # Further inspection to see if there are any other keys in the dict.
                # If not, then remove the whole item otherwise recurse
                if len(element[decode_list_index(keys[0])].keys()) == 1:
                    element.pop(decode_list_index(keys[0]))
                    return

                return nested_delete(element[decode_list_index(keys[0])], keys[1:])
        elif isinstance(keys[0], str) and isinstance(keys[1], bytes):
            return nested_delete(element[keys[0]], keys[1:])

        elif isinstance(keys[0], str) and isinstance(keys[1], str):
            #
            if isinstance(element[keys[0]], dict):
                # Case where we have an embedded
                return nested_delete(element[keys[0]], keys[1:])
            else:
                element[keys[0]] = None
                return

    if isinstance(keys[0], bytes):
        return nested_delete(element[decode_list_index(keys[0])], keys[1:])

    return nested_delete(element[keys[0]], keys[1:])


def nested_get(element, keys):
    """
    Getter for dictionary / list elements based on a key_path.

    :param element: A generic dictionary or list
    :param keys: List of string and byte encoded integers.
    :return: The value from the key_path
    """
    num_elements = len(keys)

    if num_elements == 0:
        return element

    if num_elements == 1:
        if isinstance(keys[0], bytes):
            return element[decode_list_index(keys[0])]
        return element[keys[0]]

    if isinstance(keys[0], bytes):
        return nested_get(element[decode_list_index(keys[0])], keys[1:])
    return nested_get(element[keys[0]], keys[1:])


def nested_set(element, keys, value, index: int = 0):
    """
    Set the value of an arbitrary object based on a key_path in the form of a list
    with strings for keys and byte encoded integers for indexes in a list. This function
    recurses through the element until it is at the end of the keys where it sets it.

    :param element: A generic dictionary or list
    :param keys: List of string and byte encoded integers.
    :param value: The value to set
    :param index: Index is only used when called recursively, not initially
    """
    num_elements = len(keys)
    # Check if we are at the last element of the list to insert the value
    if index == num_elements - 1:
        if isinstance(keys[-1], bytes):
            element.insert(decode_list_index(keys[-1]), value)
        else:
            element[keys[-1]] = value
        return

    if isinstance(element, dict):
        # Look ahead if the next item is a list
        if isinstance(keys[index + 1], bytes):
            element = element.setdefault(keys[index], [])
        else:
            element = element.setdefault(keys[index], {})

    # element must be list
    # Check if we are updating an existing element
    elif len(element) > decode_list_index(keys[index]):
        element = element[decode_list_index(keys[index])]
    else:
        if isinstance(keys[index + 1], bytes):
            element.insert(decode_list_index(keys[index]), [])
        else:
            element.insert(decode_list_index(keys[index]), {})
        element = element[decode_list_index(keys[index])]
    nested_set(element, keys, value, index + 1)


def get_target_and_key(context: 'Context', key_path: list = None) -> (Any, list):
    target_context = context.public_context

    if key_path is None:
        key_path = context.key_path

    output_key_path = []
    for i in key_path:
        if i == '_>':
            if context.private_context is None:
                context.private_context = (
                    {} if isinstance(output_key_path[0], str) else []
                )
            target_context = context.private_context
        elif i != '->':
            output_key_path.append(i)

    return target_context, output_key_path


def smush_key_path(key_path: list) -> list:
    """Remove the arrows from a key path."""
    output = []
    for i in key_path:
        if i not in ('->', '_>'):
            output.append(i)
    return output


def set_key(
    context: 'Context',
    value: Any,
    key_path: list = None,
    append_hook_value: bool = False,
):
    """
    Wrap nested_set to set keys for both public and private hook calls.

    For public hook calls, qualifies if the hook is compact form (ie key->) or expanded
    (ie key: {->:..}) before setting the output. For private hook calls, the key and
    all parent keys without additional objects are deleted later as they might be
    used in rendering so they are added as well but their key paths are tracked for
    later deletion.
    """
    if key_path is None:
        key_path = context.key_path

    target_context, set_key_path = get_target_and_key(context, key_path=key_path)
    nested_set(target_context, set_key_path, value)

    if len(context.key_path_block) != 0:
        tmp_key_path = key_path[(len(context.key_path_block) - len(key_path)) :]
        if context.temporary_context is None:
            context.temporary_context = {} if isinstance(tmp_key_path[0], str) else []
        tmp_key_path = [i for i in tmp_key_path if i not in ('->', '_>')]

        if tmp_key_path:
            # Assert that the list is not empty - handles cases where we are appending
            #  a value from a list.
            nested_set(context.temporary_context, tmp_key_path, value)


def _clean_item(element: Union[dict, list], item: Union[int, str], value: Any):
    if isinstance(value, dict):
        try:
            value_key = next(iter(value.keys()))
        except StopIteration:
            return
        if isinstance(value_key, CommentedKeyMap):
            new_key = next(iter(value_key.keys()))
            element[item] = "{{" + new_key + "}}"
        else:
            cleanup_unquoted_strings(value)
    elif isinstance(value, list):
        cleanup_unquoted_strings(value)


def cleanup_unquoted_strings(element: Union[dict, list]):
    """
    Cleanup a complex dict that might have unquoted templates in them which is a very
     common mistake to make when writing tackle files. Dicts could be complex with both
     lists and embedded dicts so each key needs to be visited to check. Looks for values
     like `ordereddict([('foo', None)]): None and turns them into `foo`. Input is always
     a dict.
    """
    if isinstance(element, dict):
        for k, v in element.items():
            _clean_item(element, k, v)
    elif isinstance(element, list):
        for i, v in enumerate(element):
            _clean_item(element, i, v)


def merge(a, b, path=None, update=True):
    """
    See https://stackoverflow.com/questions/7204805/python-dictionaries-of-dictionaries-merge
    Merges b into a
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            elif isinstance(a[key], list) and isinstance(b[key], list):
                for idx, val in enumerate(b[key]):
                    a[key][idx] = merge(
                        a[key][idx],
                        b[key][idx],
                        path + [str(key), str(idx)],
                        update=update,
                    )
            elif update:
                a[key] = b[key]
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a



import os

from tackle.models import BaseHook, Field


class PathExistsListHook(BaseHook):
    """Hook for os package 'path.exists'."""

    hook_type: str = 'path_exists'
    path: str = Field(..., description="The path to file or directory")

    args: list = ['path']

    def exec(self) -> bool:
        return os.path.exists(self.path)


class PathIsDirListHook(BaseHook):
    """Hook for os package 'path.isdir'."""

    hook_type: str = 'isdir'
    path: str = Field(..., description="The path to a directory")

    args: list = ['path']

    def exec(self) -> bool:
        return os.path.isdir(self.path)


class PathIsFileListHook(BaseHook):
    """Hook for os package 'path.isfile'."""

    hook_type: str = 'isfile'
    path: str = Field(..., description="The path to a file")

    args: list = ['path']

    def exec(self) -> bool:
        return os.path.isfile(self.path)


class PathJoinHook(BaseHook):
    """Hook joining paths."""

    hook_type: str = 'path_join'
    paths: list = Field(
        ...,
        description="List of items in a path to file or directory.",
        render_by_default=True,
    )

    args: list = ['paths']

    def exec(self):
        return os.path.join(*self.paths)


class PathBasenameHook(BaseHook):
    """Hook for getting the basename from a path."""

    hook_type: str = 'basename'
    path: str = Field(
        ...,
        description="Path to the file/directory to get the basename of.",
        render_by_default=True,
    )

    args: list = ['path']

    def exec(self):
        return os.path.basename(self.path)


class PathDirNameHook(BaseHook):
    """Hook for getting the basename from a path."""

    hook_type: str = 'dirname'
    path: str = Field(
        ...,
        description="Path to the file/directory to get the directory name of.",
        render_by_default=True,
    )

    args: list = ['path']

    def exec(self):
        return os.path.dirname(self.path)



from tackle import BaseHook, Field
from tackle.parser import get_hook


class FlattenHook(BaseHook):
    """
    Hook for flattening args/kwargs/flags into CLI inputs. Takes a declarative hook
    input and turns it into a string that can be used to call a generic CLI program.
    """

    hook_type: str = 'flatten'
    hook: str = Field(None, description="A hook")

    positional_args: list = Field(
        None, description="A list of args that should be considered positional."
    )
    inputs: dict = Field(
        None, description="A dictionary of arguments to use when flattening."
    )
    kwargs: str = 'inputs'
    args: list = ['hook', 'positional_args']

    def exec(self) -> str:
        if self.inputs is None:
            self.inputs: dict = self.existing_context

        hook = get_hook(context=self, hook_type=self.hook)
        if hook is None:
            raise Exception(f"The hook={self.hook} was not found.")

        flat_items = []
        for i in hook.__fields__['args'].default:
            hook_value = self.inputs[i]
            # TODO: Implment a skip or something?
            # hook.__fields__[i].default
            flat_items.append(hook_value)

        for i in hook.__fields__['function_fields'].default:
            hook_field = hook.__fields__[i]
            hook_value = self.inputs[i]

            if hook_field.type_ == bool:
                if hook_field.name in self.inputs and self.inputs[hook_field.name]:
                    flat_items.append("--" + hook_field.name)
                elif hook_field.default:
                    flat_items.append("--" + hook_field.name)
            elif hook_field.type_ in (str, int, float):
                if hook_field.name in hook.__fields__['args'].default:
                    continue
                if hook_value == hook.__fields__[hook_field.name].default:
                    continue
                flat_items.append("--" + hook_field.name)
                flat_items.append(hook_value)
            else:
                raise NotImplementedError(
                    "No other meanings to types have been determined"
                )
            pass

        return ' '.join(flat_items)



import pytest

from tackle import tackle
from tackle.models import Context
from tackle.parser import update_source

FIXTURES = [
    'expanded-string.yaml',
    'expanded-list-dict.yaml',
    'local.yaml',
    'function-import.yaml',
]


@pytest.mark.parametrize("target", FIXTURES)
def test_provider_system_hook_import(change_dir, target):
    """Run the source and check that the hooks imported the demo module."""
    context = Context(input_string=target)
    # num_providers = len(context.provider_hooks.keys())
    num_providers = len(context.private_hooks.keys())
    update_source(context)
    # assert num_providers < len(context.provider_hooks.keys())
    assert num_providers < len(context.private_hooks.keys())
    # assert 'tackle.providers.tackle-demos' in context.providers.hook_types


def test_provider_system_hook_import_local(change_dir):
    """Assert local import of hook is valid."""
    o = tackle('local.yaml')
    assert o['stuff'] == 'thing'



"""Tests dict input objects for `tackle.providers.system.hooks.listdir` module."""
from tackle.main import tackle


def test_provider_system_hook_listdir(change_dir):
    output = tackle('.', no_input=True)

    assert len(output['string_input']) == 3
    assert len(output['string_input_sorted']) == 2
    assert len(output['list_input']) == 2



from tackle import tackle


def test_hook_sort_in_place(change_dir):
    output = tackle('sort-in-place.yaml')
    assert output['a_list'] == ['bar', 'baz', 'foo']
    assert output['b_list'] == ['foo', 'bar', 'baz']
    assert output['sort_in_list'] == ['bar', 'baz', 'foo']


def test_hook_sort_map_key(change_dir):
    output = tackle('sort-map.yaml')
    assert output['sort_a_map'][0]['foo'] == 'bar'


def test_collections_hook_list_key(change_dir):
    output = tackle('list-keys.yaml')
    assert output['list_key_values'] == ['things', 'mo tings']


def test_collections_hook_concatenate(change_dir):
    output = tackle('concatenate.yaml')
    assert output['out'] == ['foo', 'bar', 'stuff', 'things']


def test_collections_hook_distinct(change_dir):
    output = tackle('distinct.yaml')
    for i in output['distincts']:
        assert i in ['stuff', 'foo', 'things']



"""Tests for tackle.utils.dicts"""
import pytest
from ruamel.yaml import YAML

from tackle.utils.dicts import (
    encode_list_index,
    decode_list_index,
    nested_set,
    nested_get,
    nested_delete,
    # set_key,
    cleanup_unquoted_strings,
)

ZERO_INDEX = encode_list_index(0)


def test_index_encoding():
    """Test encoding function."""
    assert decode_list_index(encode_list_index(1)) == 1


NESTED_SET_FIXTURES = [
    # Indexed as output, key_path, expected_output
    ({}, ['foo'], {'foo': True}),
    ({}, ['one', 'two', 'three'], {'one': {'two': {'three': True}}}),
    ({}, ['this', 'lists', ZERO_INDEX], {'this': {'lists': [True]}}),
    (
        {},
        ['this', 'lists', ZERO_INDEX, 'one', 'two'],
        {'this': {'lists': [{'one': {'two': True}}]}},
    ),
    (
        {},
        ['this', 'lists', ZERO_INDEX, 'one', 'two', ZERO_INDEX],
        {'this': {'lists': [{'one': {'two': [True]}}]}},
    ),
    (
        {},
        ['this', 'lists', ZERO_INDEX, 'one', 'two', ZERO_INDEX, 'one', 'two', 'three'],
        {'this': {'lists': [{'one': {'two': [{'one': {'two': {'three': True}}}]}}]}},
    ),
    ({'bar': 2}, ['foo'], {'bar': 2, 'foo': True}),
    ({'bar': 2}, ['one', 'two', 'three'], {'bar': 2, 'one': {'two': {'three': True}}}),
    ({'bar': 2}, ['this', 'lists', ZERO_INDEX], {'bar': 2, 'this': {'lists': [True]}}),
    (
        {},
        ['this', ZERO_INDEX, 'one', 'two', ZERO_INDEX],
        {'this': [{'one': {'two': [True]}}]},
    ),
    ({}, ['this', ZERO_INDEX, 'one'], {'this': [{'one': True}]}),
    (
        {},
        ['this', 'that', 'foo', ZERO_INDEX, 'one'],
        {'this': {'that': {'foo': [{'one': True}]}}},
    ),
    (
        {'this': {'that': {'foo': [{'one': 1}]}}},
        ['this', 'that', 'foo', encode_list_index(1), 'two'],
        {'this': {'that': {'foo': [{'one': 1}, {'two': True}]}}},
    ),
    (
        {'this': {'lists': [{'one': {'two': 1}}]}},
        ['this', 'lists', ZERO_INDEX, 'one', 'three'],
        {'this': {'lists': [{'one': {'two': 1, 'three': True}}]}},
    ),
    (
        {'this': {'lists': [{'one': {'two': 1}}]}},
        ['this', 'lists', encode_list_index(1), 'one', 'three'],
        {'this': {'lists': [{'one': {'two': 1}}, {'one': {'three': True}}]}},
    ),
    (
        {'foo': ['bar'], 'this': [{'foo': []}]},
        ['this', encode_list_index(0), 'foo', encode_list_index(0), 'one'],
        {'foo': ['bar'], 'this': [{'foo': [{'one': True}]}]},
    ),
    (
        {'this': {'lists': [{'one': {'two': 1}}]}},
        ['this', 'lists', encode_list_index(1), 'one', 'three'],
        {'this': {'lists': [{'one': {'two': 1}}, {'one': {'three': True}}]}},
    ),
    ({}, ['this', ZERO_INDEX, 'one', ZERO_INDEX], {'this': [{'one': [True]}]}),
    (
        {},
        ['this', encode_list_index(0), 'foo', encode_list_index(0), 'one'],
        {'this': [{'foo': [{'one': True}]}]},
    ),
    (
        {},
        ['this', encode_list_index(0), 'foo', 'one'],
        {'this': [{'foo': {'one': True}}]},
    ),
    (
        {},
        ['this', ZERO_INDEX, ZERO_INDEX, 'foo'],
        {'this': [[{'foo': True}]]},
    ),
    (
        {'stuff': ['things']},
        ['stuff', encode_list_index(1), encode_list_index(0)],
        {'stuff': ['things', [True]]},
    ),
    (
        [{'stuff': False}],
        [encode_list_index(1), 'stuff', encode_list_index(0)],
        [{'stuff': False}, {'stuff': [True]}],
    ),
    (
        [],
        [encode_list_index(0), 'stuff', encode_list_index(0)],
        [{'stuff': [True]}],
    ),
    (
        [],
        [encode_list_index(0), 'stuff', encode_list_index(0)],
        [{'stuff': [True]}],
    ),
    # # This test is breaking
    # (
    #     {'stuff': ['things', [{'foo': ['bar', ['baz']]}]]},
    #     ['stuff', encode_list_index(1), encode_list_index(0), 'foo', encode_list_index(1)],
    #     {'stuff': ['things', [{'foo': ['bar', True]}]]},
    # ),
]


@pytest.mark.parametrize("output,key_path,expected_output", NESTED_SET_FIXTURES)
def test_nested_append_bytes(output, key_path, expected_output):
    """Test setting elements in list."""
    nested_set(output, key_path, True)
    assert output == expected_output


NESTED_GET_FIXTURES = [
    ({'foo': 1}, ['foo'], 1),
    (
        {'this': {'that': {'foo': [{'one': 1}, {'one': 'foo'}]}}},
        ['this', 'that', 'foo', encode_list_index(1), 'one'],
        'foo',
    ),
    ({'list': [{'->': 'var this'}]}, ['list', ZERO_INDEX, '->'], 'var this'),
    (
        {'list': [{'->': 'var this'}, {'key->': 'var this'}]},
        ['list', b'\x00\x01', 'key->'],
        'var this',
    ),
    ({'foo': [1]}, ['foo', ZERO_INDEX], 1),
]


@pytest.mark.parametrize("input,key_path,expected_output", NESTED_GET_FIXTURES)
def test_nested_get(input, key_path, expected_output):
    """Test getting based on key path."""
    output = nested_get(input, key_path)

    assert output == expected_output


NESTED_DELETE_FIXTURES = [
    ({'foo': 1}, ['foo'], {}),
    ({'foo': 1, 'bar': 1}, ['foo'], {'bar': 1}),
    ({'foo': 'baz', 'bar': 1}, ['foo', 'baz'], {'foo': None, 'bar': 1}),
    ({'list': [{'->': 'var this'}]}, ['list', ZERO_INDEX, '->'], {'list': []}),
    (
        {'list': [{'->': 'var this'}, 'foo']},
        ['list', ZERO_INDEX, '->'],
        {'list': ['foo']},
    ),
    (
        {'this': {'that': {'foo': [{'one': 1}, {'one': 'foo'}]}}},
        ['this', 'that', 'foo', encode_list_index(1), 'one'],
        {'this': {'that': {'foo': [{'one': 1}]}}},
    ),
    (
        {'this': {'that': {'foo': [{'one': 1}, {'one': 'foo'}]}}},
        ['this', 'that', 'foo', encode_list_index(1)],
        {'this': {'that': {'foo': [{'one': 1}]}}},
    ),
    (
        {'this': {'that': {'foo': [{'one': 1}, {'one': 'foo', 'two': 'boo'}]}}},
        ['this', 'that', 'foo', encode_list_index(1), 'one'],
        {'this': {'that': {'foo': [{'one': 1}, {'two': 'boo'}]}}},
    ),
    ({'foo': {'baz': 1, 'bar': 1}}, ['foo', 'bar'], {'foo': {'baz': 1}}),
    ({'foo': 'bar', 'baz': 1}, ['foo', 'bar'], {'foo': None, 'baz': 1}),
    ({'foo': {'bar': 1, 'baz': 1}}, ['foo', 'bar'], {'foo': {'baz': 1}}),
    ({'foo': {'bar': 1}, 'baz': 1}, ['foo', 'bar'], {'foo': {}, 'baz': 1}),
    ({'foo': 'bar'}, ['foo', 'bar'], {'foo': None}),
    (
        {'list': [{'->': 'var this'}]},
        ['list', ZERO_INDEX, '->', 'var this'],
        {'list': [{'->': None}]},
    ),
    (
        {'list': ['foo', ['a', 'embedded', 'list']]},
        ['list', encode_list_index(1)],
        {'list': ['foo']},
    ),
    (
        {'list': ['foo', ['a', 'embedded', 'list']]},
        ['list', encode_list_index(1), encode_list_index(0)],
        {'list': ['foo', ['embedded', 'list']]},
    ),
]


@pytest.mark.parametrize("input,key_path,expected_output", NESTED_DELETE_FIXTURES)
def test_nested_delete(input, key_path, expected_output):
    """Test deletion based on key path."""
    nested_delete(input, key_path)
    assert input == expected_output


CLEANUP_UNQUOTED_STRINGS_FIXTURES = [
    (
        """
        foo: bar
        """,
        {'foo': 'bar'},
    ),
    (
        """
        foo: {{bar}}
        """,
        {'foo': '{{bar}}'},
    ),
    (
        """
        foo:
          bar: {{baz}}
        """,
        {'foo': {'bar': '{{baz}}'}},
    ),
    (
        """
        foo:
          bar:
            - {{baz}}
        """,
        {'foo': {'bar': ['{{baz}}']}},
    ),
    (
        """
        foo:
          bar:
            - baz: 1
              bing: {{baz}}
        """,
        {'foo': {'bar': [{'baz': 1, 'bing': '{{baz}}'}]}},
    ),
    (
        """
        foo:
          bar:
            - baz: 1
              bing: baz
        """,
        {'foo': {'bar': [{'baz': 1, 'bing': 'baz'}]}},
    ),
]


@pytest.mark.parametrize("input,expected_output", CLEANUP_UNQUOTED_STRINGS_FIXTURES)
def test_utils_dicts_cleanup_unquoted_strings(input, expected_output):
    """Test deletion based on key path."""
    yaml = YAML()
    input_dict = yaml.load(input)

    cleanup_unquoted_strings(input_dict)
    assert input_dict == expected_output



"""Path related utils."""
import contextlib
import errno
import os
import shutil
import stat
import logging
import re
from typing import Optional

from tackle.exceptions import ContextFileNotFound, InvalidConfiguration

logger = logging.getLogger(__name__)

CONTEXT_FILES = {
    '.tackle.yaml',
    '.tackle.yml',
    '.tackle.json',
    'tackle.yaml',
    'tackle.yml',
    'tackle.json',
}


def listdir_absolute(directory, skip_paths=None):
    """Return and iterator of the absolute path."""
    if skip_paths is None:
        skip_paths = []
    for dirpath, _, filenames in os.walk(directory):
        for f in filenames:
            if f not in skip_paths:
                yield os.path.abspath(os.path.join(dirpath, f))


def force_delete(func, path, exc_info):
    """Error handler for `shutil.rmtree()` equivalent to `rm -rf`.

    Usage: `shutil.rmtree(path, onerror=force_delete)`
    From stackoverflow.com/questions/1889597
    """
    os.chmod(path, stat.S_IWRITE)
    func(path)


def rmtree(path):
    """Remove a directory and all its contents. Like rm -rf on Unix.

    :param path: A directory path.
    """
    if os.path.islink(path):
        os.unlink(path)
    else:
        # Regular file
        shutil.rmtree(path, onerror=force_delete)


def make_sure_path_exists(path):
    """Ensure that a directory exists.

    :param path: A directory path.
    """
    logger.debug('Making sure path exists: %s', path)
    try:
        os.makedirs(path)
        logger.debug('Created directory at: %s', path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            return False
    return True


def make_executable(script_path):
    """Make `script_path` executable.

    :param script_path: The file to change
    """
    status = os.stat(script_path)
    os.chmod(script_path, status.st_mode | stat.S_IEXEC)


def expand_paths(paths: list) -> list:
    """Expand a list of paths."""
    output = []
    for i in paths:
        output.append(os.path.abspath(os.path.expanduser(os.path.expandvars(i))))
    return output


def expand_path(path: str) -> str:
    """Expand both environment variables and user home in the given path."""
    path = os.path.expandvars(path)
    path = os.path.expanduser(path)
    return path


def expand_abbreviations(template, abbreviations) -> str:
    """Expand abbreviations in a template name.

    :param template: The project template name.
    :param abbreviations: Abbreviation definitions.
    """
    if template in abbreviations:
        return abbreviations[template]

    # Split on colon. If there is no colon, rest will be empty
    # and prefix will be the whole template
    prefix, sep, rest = template.partition(':')
    if prefix in abbreviations:
        return abbreviations[prefix].format(rest)

    return template


def is_repo_url(value) -> bool:
    """
    Return True if value is a repository URL. Also checks if it is of the form
    org/repo ie robcxyz/tackle-demos and checks if that is a local directory before
    returning true.
    """
    REPO_REGEX = re.compile(
        r"""
    # something like git:// ssh:// file:// etc.
    ((((git|hg)\+)?(git|ssh|file|https?):(//)?)
     |
     (^(github.com|gitlab.com)+\/[\w,\-,\_]+\/[\w,\-,\_]+$)
    )
    """,
        re.VERBOSE,
    )
    if not bool(REPO_REGEX.match(value)):
        # We also need to catch refs of type robcxyz/tackle-demo but need to first see
        # if the path
        REPO_REGEX = re.compile(
            r"""^[\w,\-,\_]+\/[\w,\-,\_]+$""",
            re.VERBOSE,
        )
        if bool(REPO_REGEX.match(value)):
            if os.path.exists(value):
                return False
            else:
                return True
        else:
            return False

    return True
    # return bool(REPO_REGEX.match(value))


def is_directory_with_tackle(value) -> bool:
    """Return true if directory."""
    if os.path.isdir(value):
        contents = os.listdir(value)
        for i in CONTEXT_FILES:
            if i in contents:
                return True
    return False


def is_file(value) -> bool:
    """Return True if the input looks like a file."""
    FILE_REGEX = re.compile(
        r"""^.*\.(yaml|yml|json|toml)$""",
        re.VERBOSE,
    )
    return bool(FILE_REGEX.match(value))


def find_tackle_file(provider_dir) -> str:
    """Find the tackle files based on some defaults and return the path."""
    provider_contents = os.listdir(provider_dir)
    for i in provider_contents:
        if i in CONTEXT_FILES:
            return os.path.join(provider_dir, i)

    raise InvalidConfiguration(f"Can't find tackle file in {provider_dir}")


def find_in_parent(dir: str, targets: list, fallback=None) -> str:
    """Recursively search in parent directories for a path to a target file."""
    for i in os.listdir(dir):
        if i in targets:
            return os.path.abspath(os.path.join(dir, i))

    if os.path.abspath(dir) == '/':
        if fallback:
            return fallback
        else:
            raise NotADirectoryError(
                f'The {targets} target doesn\'t exist in the parent directories.'
            )
    return find_in_parent(
        dir=os.path.dirname(os.path.abspath(dir)),
        targets=targets,
        fallback=fallback,
    )


def find_nearest_tackle_file() -> Optional[str]:
    """
    Find the nearest tackle file from a set of default tackle files.
    :return: Path or None if not found
    """
    tackle_file_location = find_in_parent(os.curdir, CONTEXT_FILES, fallback=True)
    if isinstance(tackle_file_location, bool):
        return None

    return tackle_file_location


def repository_has_tackle_file(repo_directory: str, context_file=None):
    """
    Determine if `repo_directory` contains a valid context file. Acceptable choices
    are `tackle.yaml`, `tackle.yml`, `tackle.json`, `cookiecutter.json` in order of
    precedence.

    :param repo_directory: The candidate repository directory.
    :param context_file: eg. `tackle.yaml`.
    :return: The path to the context file
    """
    if context_file:
        # The supplied context file exists
        context_file = os.path.join(os.path.abspath(repo_directory), context_file)
        if os.path.isfile(context_file):
            return context_file
        else:
            raise ContextFileNotFound(
                f"Can't find supplied context_file at {context_file}"
            )

    repo_directory_exists = os.path.isdir(repo_directory)
    if repo_directory_exists:
        # Check for valid context files as default
        for f in CONTEXT_FILES:
            if os.path.isfile(os.path.join(repo_directory, f)):
                return f
    else:
        return None


@contextlib.contextmanager
def work_in(dirname=None):
    """Context manager version of os.chdir.

    When exited, returns to the working directory prior to entering.
    """
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)



import pytest
from tackle.cli import main

COMMANDS = [
    (['global-kwarg.yaml', '--key_a', '"stuff and things"'], 'stuff and things'),
]


@pytest.mark.parametrize("command,expected_output", COMMANDS)
def test_cli_commands(change_curdir_fixtures, command, expected_output, capsys):
    """Assert output comes out of cli."""
    main(command)
    assert expected_output in capsys.readouterr().out


def test_cli_command_find_in_parent(chdir_fixture, capsys):
    """Check that we can change into a child dir and find a file in parent with flag."""
    chdir_fixture('dir')
    main(['global-kwarg.yaml', '--key_a', '"stuff and things"', '--find-in-parent'])
    assert 'stuff and things' in capsys.readouterr().out



from git import Repo
from typing import Optional

from tackle.models import BaseHook, Field


class GitCloneHook(BaseHook):
    """
    Hook to create clone a repo. Wraps git python `clone`.
    [Source API](https://gitpython.readthedocs.io/en/stable/reference.html#git.repo.base.Repo.clone)
    """

    hook_type: str = 'git_clone'
    url: str = Field(
        None,
        description="valid git url - [Docs](http://www.kernel.org/pub/software/scm/git/docs/git-clone.html#URLS)",
    )
    to_path: str = Field(
        None, description="Path to which the repository should be cloned to"
    )
    progress: Optional[str] = Field(None, description="See ‘git.remote.Remote.push’.")
    env: Optional[dict] = Field(
        None,
        description="Optional dictionary containing the desired environment "
        "variables. Note: Provided variables will be used to update "
        "the execution environment for git. If some variable is not "
        "specified in env and is defined in os.environ, value from "
        "os.environ will be used. If you want to unset some "
        "variable, consider providing empty string as its value.",
    )
    multi_options: Optional[list] = Field(
        None,
        description="A list of Clone options that can be provided multiple "
        "times. One option per list item which is passed exactly as "
        "specified to clone. For example "
        "[‘–config core.filemode=false’, ‘–config core.ignorecase’, "
        "‘–recurse-submodule=repo1_path’, "
        "‘–recurse-submodule=repo2_path’]",
    )
    kwargs: Optional[dict] = Field(
        {},
        description="- odbt = ObjectDatabase Type, allowing to determine the "
        "object database implementation used by the returned Repo "
        "instance "
        "- All remaining keyword arguments are given to the git-clone "
        "command",
    )

    args: list = ['url', 'to_path']

    def exec(self):
        if not self.to_path:
            self.to_path = self.url.split('/')[-1].split('.')[0]
        if self.url:
            Repo.clone_from(
                url=self.url,
                to_path=self.to_path,
                progress=self.progress,
                env=self.env,
                multi_options=self.multi_options,
                **self.kwargs,
            )
            return self.to_path



"""Example."""
from tackle.models import BaseHook
from pydantic import validator


class ThingHook(BaseHook):
    """Do stuff and things."""

    hook_type: str = 'thing'
    stuff: str

    @validator('stuff')
    def validate(cls, value):
        # Check if the input is valid - throw error otherwise
        if value == 'not-things':
            raise Exception
        return value

    def exec(self):
        print(self.stuff)
        return self.stuff



"""Data type hook."""
from typing import Any
from pydantic import Field
from tackle.models import BaseHook


class TypeHook(BaseHook):
    """Hook for getting the type of a variable."""

    hook_type: str = 'type'

    input: Any = Field(..., description="Any variable input.", render_by_default=True)

    args: list = ['input']

    def exec(self) -> str:
        output = type(self.input).__name__
        if output == 'CommentedSeq':
            output = 'list'
        elif output == 'CommentedMap':
            output = 'map'
        return output



from tackle.main import tackle


def test_provider_paths_flatten(change_dir):
    output = tackle()
    assert output['flattened'][0] == 'one/baz/stuff.yaml'
    assert output['flattened'][9] == 'complex/alist/another_list/things.py'
    assert output['flattened_base'][0] == 'base/path/one/baz/stuff.yaml'
    assert output['flattened_base_list'][0] == 'base/path/foo/bar.py'



"""Assert hook."""
from typing import Any

from tackle.models import BaseHook, Field
from tackle import exceptions


class AssertHook(BaseHook):
    """Hook for asserting an input is equal to a value."""

    hook_type: str = 'assert'
    input: Any = Field(..., description="The input - ie right hand side of assert.")
    value: Any = Field(None, description="The value - ie left hand side of assert.")
    exit_on_failure: bool = Field(
        True, description="Whether to exit on assertion error."
    )

    args: list = ['input', 'value']
    _docs_order = 5

    def exec(self) -> bool:
        """Runs for both if the value is present or full assertion."""
        if self.value is None:
            if not self.exit_on_failure:
                try:
                    assert self.input
                    return True
                except AssertionError:
                    return False
            else:
                try:
                    assert self.input
                    return True
                except AssertionError:
                    raise exceptions.HookCallException(
                        f"Error asserting {self.input}=={self.value}", context=self
                    ) from None
        else:
            if not self.exit_on_failure:
                try:
                    assert self.input == self.value
                    return True
                except AssertionError:
                    return False
            else:
                try:
                    assert self.input == self.value
                    return True
                except AssertionError:
                    raise exceptions.HookCallException(
                        f"Error asserting {self.input}=={self.value}", context=self
                    ) from None



# from tackle.main import tackle
#
# # TODO: https://github.com/robcxyz/tackle-box/issues/46
#
# def test_provider_pyinquirer_input_hook(change_dir, mocker):
#     # mocker.patch(
#     #     # 'tackle.providers.prompts.hooks.input.prompt',
#     #     'tackle.main.update_source',
#     #     return_value={"tmp": True},
#     #     autospec=True
#     # )
#     o = tackle()
#     assert o['this']



import os
import random
import string
from abc import ABC
from pydantic import (
    BaseModel,
    SecretStr,
    Field,
    Extra,
    validator,
)
from pydantic.main import ModelMetaclass
from jinja2 import Environment, StrictUndefined
from jinja2.ext import Extension
from typing import Any, Union, Optional, Callable

from tackle.hooks import import_native_providers, import_hooks_from_dir
from tackle.utils.render import wrap_jinja_braces
from tackle.exceptions import TooManyTemplateArgsException


class StrictEnvironment(Environment):
    """Create strict Jinja2 environment.

    Jinja2 environment will raise error on undefined variable in template-
    rendering context.
    """

    def __init__(self, **kwargs):
        super(StrictEnvironment, self).__init__(undefined=StrictUndefined, **kwargs)
        # TODO: Add imports for jinja hook filters


class BaseContext(BaseModel):
    """Shared items that are passed between hook calls and the main context."""

    no_input: bool = False
    verbose: bool = False

    input_context: Union[dict, list] = None
    public_context: Union[dict, list] = None
    private_context: Union[dict, list] = None
    temporary_context: Union[dict, list] = None
    existing_context: dict = {}

    key_path: list = []
    key_path_block: list = []

    # public_hooks: dict = {}
    public_hooks: dict = None
    private_hooks: dict = None
    default_hook: Any = Field(None, exclude=True)

    calling_directory: str = None
    calling_file: str = None
    current_file: str = None

    env_: Any = None

    class Config:
        smart_union = True

    # TODO: RM after dealing with namespace issue for validators
    # https://github.com/robcxyz/tackle-box/issues/43
    run_id: str = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.run_id is None:
            self.run_id: str = ''.join(
                random.choice(string.ascii_uppercase + string.digits) for _ in range(4)
            )


class Context(BaseContext):
    """The main object that is being modified by parsing."""

    # Source
    password: SecretStr = None
    directory: str = None
    find_in_parent: bool = False
    input_string: str = None
    input_dir: str = None
    input_file: str = None
    override_context: Union[str, dict] = Field(
        None, description="A str for a file or dict to override inputs with."
    )

    context_functions: list = None

    hook_dirs: list = Field(
        None, description="A list of additional directories to import hooks."
    )

    latest: bool = False
    # TODO: Change to version?
    checkout: str = Field(
        None,
        description="For inputs referencing remote repos, refers to a branch or tag.",
    )

    global_args: list = None
    global_kwargs: dict = None
    global_flags: list = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.latest:
            self.checkout = 'latest'

        # Allows for passing the providers between tackle runtimes
        if self.private_hooks is None:
            self.private_hooks = {}
            import_native_providers(self)

        if self.public_hooks is None:
            self.public_hooks = {}

        if self.hook_dirs is not None:
            for i in self.hook_dirs:
                import_hooks_from_dir(context=self, mod_name=i, path=i)

        if self.env_ is None:
            self.env_ = StrictEnvironment()

        if self.calling_directory is None:
            # Can be carried over from another context. Should only be initialized when
            # called from CLI.
            self.calling_directory = os.path.abspath(os.path.curdir)


class BaseHook(BaseContext, Extension):
    """
    Base hook class from which all other hooks inherit from to be discovered. There are
    a number of reserved keys that are used for logic such as `if` and `for` that are
    aliased to `if_` and `for_` to not collide with python reserved keywords.
    """

    hook_type: str = Field(..., description="Name of the hook.")

    if_: Union[str, bool] = Field(None, render_by_default=True)
    when: Union[str, bool] = Field(None, render_by_default=True)
    else_: Any = Field(None)
    for_: Union[str, list] = Field(None, render_by_default=True)
    reverse: Union[str, bool] = Field(None, render_by_default=True)
    try_: Union[str, bool] = Field(None, render_by_default=True)
    except_: Any = Field(None, render_by_default=True)
    chdir: Optional[str] = Field(None, description="Name of the hook.")
    merge: Union[bool, str] = None
    confirm: Optional[Any] = None

    # Flag for whether being called directly (True) or by a jinja extension (False)
    is_hook_call: bool = False
    skip_import: bool = False
    hooks_path: str = None
    is_public: bool = False

    env_: Any = None

    args: list = []
    kwargs: Union[str, dict] = None

    skip_output: bool = False
    # Fields that should not be rendered by default
    _render_exclude_default: set = {
        'input_context',
        'public_context',
        'hook_type',
        'else',
    }
    _render_exclude: set = {}
    _render_by_default: list = []

    # Used when rendering docs
    _doc_tags: list = []
    # For linking issues in the docs so others can potentially contribute
    _issue_numbers: list = []  # TODO: Implement this
    # Additional callout sections to be included at the top of the docs
    _notes: list = []
    # Allow hooks to be sorted in the docs
    _docs_order: int = 10  # Arbitrary high number so hooks can be sorted high or low
    # Parameterized return type description for docs
    _return_description: str = None
    # Flag for skipping creating the docs for
    _wip: bool = False

    @validator('if_', 'reverse', 'for_', 'merge')
    def wrap_bool_if_string(cls, v):
        return wrap_jinja_braces(v)

    # Per https://github.com/samuelcolvin/pydantic/issues/1577
    def __setattr__(self, key, val):
        """Override method to alias input fields."""
        if key in self.__config__.alias_to_fields:
            key = self.__config__.alias_to_fields[key]
        super().__setattr__(key, val)

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid
        validate_assignment = True
        fields = {
            'if_': 'if',
            'else_': 'else',
            'for_': 'for',
            'try_': 'try',
            'except_': 'except',
        }
        # Per https://github.com/samuelcolvin/pydantic/issues/1577
        # This is an issue until pydantic 1.9 is released and items can be set with
        # properties which will override the internal __setattr__ method that
        # disregards aliased fields
        alias_to_fields = {v: k for k, v in fields.items()}

    # TODO: RM?
    def exec(self) -> Any:
        raise NotImplementedError("Every hook needs an exec method.")


class FunctionInput(BaseModel):
    """Function input model. Used to validate the raw function input."""

    exec_: Any = Field(None)
    return_: Union[str, list] = Field(None)

    args: list = None
    kwargs: str = None

    render_exclude: list = None
    validators: dict = None
    methods: dict = None
    public: bool = None
    extends: str = None
    help: str = None

    class Config:
        extra = 'ignore'
        validate_on_assignment = True


class BaseFunction(BaseHook, FunctionInput, ABC):
    """Base model when creating new functions."""


class JinjaHook(BaseModel):
    """
    Model for jinja hooks. Is instantiated in render.py while rendering from unknown
    variables.
    """

    hook: ModelMetaclass
    context: BaseContext

    class Config:
        # TODO: Had to add this to get `make test` to run.
        #  https://github.com/robcxyz/tackle-box/issues/90
        arbitrary_types_allowed = True

    def wrapped_exec(self, *args, **kwargs):
        # Map args / kwargs when called via a jinja global
        from tackle.parser import evaluate_args

        args_list = list(args)
        for i in args_list:
            if isinstance(i, StrictUndefined):
                raise TooManyTemplateArgsException(
                    "Too many arguments supplied to hook call", context=self.context
                )
        evaluate_args(
            args=args_list, hook_dict=kwargs, Hook=self.hook, context=self.context
        )
        # Can't simply do self.hook(**kwargs, **self.context.dict()) because the
        #  references might not be copied over properly.
        output = self.hook(
            **kwargs,
            input_context=self.context.input_context,
            public_context=self.context.public_context,
            existing_context=self.context.existing_context,
            no_input=self.context.no_input,
            calling_directory=self.context.calling_directory,
            calling_file=self.context.calling_file,
            private_hooks=self.context.private_hooks,
            public_hooks=self.context.public_hooks,
            key_path=self.context.key_path,
            verbose=self.context.verbose,
        ).exec()
        return output

    def set_method(self, key: str, method: Callable):
        """
        Method to attach method attributes onto wrapped_exec method so that function
        can be called from jinja globals. For instance JinjaHook().wrapped_exec will be
        put into jinja.environment.globals which might have a method itself such as
        JinjaHook().wrapped_exec.JinjaHook().wrapped_exec.another_method. This can't
        be done directly with setattr against JinjaHook().wrapped_exec. Inspired by
        https://stackoverflow.com/a/20306101/12642712
        """
        setattr(JinjaHook.wrapped_exec, key, method)



"""Tests dict input objects for `tackle.providers.system.hooks.copy` module."""
import os
import shutil
from tackle.main import tackle
import pytest


@pytest.fixture()
def clean_files():
    """Clean the run."""
    yield
    files = ['thing.yaml', 'thing2.yaml', 'stuff']
    for f in files:
        if os.path.isfile(f):
            os.remove(f)
        if os.path.isdir(f):
            shutil.rmtree(f)


def test_provider_system_hook_file(change_dir, clean_files):
    tackle('tackle.yaml')
    assert 'thing.yaml' in os.listdir()
    assert 'stuff' in os.listdir()
    # If the file has been moved properly there should be only one file
    assert len(os.listdir('stuff')) == 3


def test_provider_system_hook_file_shred(change_dir, clean_files):
    files = ['stuff', 'thing', 'foo']
    for f in files:
        file = open(f, "w")
        file.write(f)
        file.close()

    tackle('shred.yaml')

    for f in files:
        assert not os.path.isfile(f)


@pytest.fixture()
def fix_file_perms():
    """Change back file perms."""
    yield
    if os.name == 'nt':
        pytest.skip("Skipping when run from windows.", allow_module_level=True)
    os.chmod('tackle.yaml', int('0o644', 8))


def test_provider_system_hook_file_chmod(change_dir, fix_file_perms):
    tackle('chmod.yaml')
    assert oct(os.stat('tackle.yaml').st_mode)[-3:] == "600"


def test_provider_files_remove(change_dir, fix_file_perms):
    o = tackle('remove.yaml')
    assert o['if_file']
    assert not o['not_file']
    assert o['if_files']
    assert not o['not_files']


def test_provider_system_hook_file_file(change_dir):
    """Verify file hook can read and write."""
    o = tackle('file.yaml')
    assert os.path.exists('rm.txt')
    os.remove('rm.txt')
    assert 'command' in o['read']



from typing import TYPE_CHECKING

from tackle.utils.dicts import (
    nested_get,
    nested_delete,
    nested_set,
    get_target_and_key,
    smush_key_path,
)

from tackle import exceptions
from tackle.models import BaseHook

if TYPE_CHECKING:
    from tackle.models import Context


def var_hook_macro(args) -> list:
    """
    Handler for cases where we have a hook call with a renderable string as the first
    argument which we rewrite as a var hook. For instance `foo->: foo-{{ bar }}-baz`
    would be rewritten as `foo->: var foo-{{bar}}-baz`.
    """
    if '{{' in args[0]:
        # We split up the string before based on whitespace so eval individually
        if '}}' in args[0]:
            # This is single templatable string -> key->: "{{this}}" => args: ['this']
            args.insert(0, 'var')
        else:
            # Situation where we have key->: "{{ this }}" => args: ['{{', 'this' '}}']
            for i in range(1, len(args)):
                if '}}' in args[i]:
                    joined_template = ' '.join(args[: (i + 1)])
                    other_args = args[(i + 1) :]
                    args = ['var'] + [joined_template] + other_args
                    break
    return args


def blocks_macro(context: 'Context'):
    """
    Handle keys appended with arrows and interpret them as `block` hooks. Value is
    re-written over with a `block` hook to support the following syntax.

    a-key->:
      if: stuff == 'things'
      foo->: print ...
    to
    a-key:
      ->: block
      if: stuff == 'things'
      items:
        foo->: print ...
    """
    # Break up key paths
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]
    # Handle embedded blocks which need to have their key paths adjusted
    key_path = (base_key_path + new_key)[len(context.key_path_block) :]
    arrow = [context.key_path[-1][-2:]]

    indexed_key_path = context.key_path[
        (len(context.key_path_block) - len(context.key_path)) :
    ]
    input_dict = nested_get(
        element=context.input_context,
        keys=indexed_key_path[:-1],
    )
    value = input_dict[indexed_key_path[-1]]

    for k, v in list(input_dict.items()):
        if k == indexed_key_path[-1]:
            nested_set(context.input_context, key_path, {arrow[0]: 'block'})
            nested_delete(context.input_context, indexed_key_path)
        else:
            input_dict[k] = input_dict.pop(k)

    # Iterate through the block keys except for the reserved keys like `for` or `if`
    aliases = [v.alias for _, v in BaseHook.__fields__.items()] + ['->', '_>']
    for k, v in value.copy().items():
        if k in aliases:
            nested_set(
                element=context.input_context,
                keys=key_path + [k],
                value=v,
            )
        else:
            # Set the keys under the `items` key per the block hook's input
            nested_set(
                element=context.input_context,
                keys=key_path + ['items', k],
                value=v,
            )


def compact_hook_call_macro(context: 'Context', element: str) -> dict:
    """
    Rewrite the input_context with an expanded expression on the called compact key.
     Returns the string element as a dict with the key as the arrow and element as
     value.
    """
    # TODO: Clean this up
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]
    key_path = (base_key_path + new_key)[len(context.key_path_block) :]
    arrow = context.key_path[-1][-2:]

    extra_keys = len(context.key_path) - len(context.key_path_block)
    old_key_path = context.key_path[-extra_keys:]

    value = nested_get(
        element=context.input_context,
        keys=smush_key_path(old_key_path)[:-1],
    )

    replacement = {context.key_path[-1]: new_key[0]}
    for k, v in list(value.items()):
        value[replacement.get(k, k)] = (
            value.pop(k) if k != context.key_path[-1] else {arrow: value.pop(k)}
        )

    # Reset the key_path without arrow
    context.key_path = context.key_path_block + key_path

    return {arrow: element}


def list_to_var_macro(context: 'Context', element: list) -> dict:
    """
    Convert arrow keys with a list as the value to `var` hooks via a re-write to the
    input.
    """
    # TODO: Convert this to a block. Issue is that keys are not rendered by default so
    #  when str items in a list are parsed, they are not rendered by default. Should
    #  have some validator or something on block to render str items in a list.
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]
    old_key_path = context.key_path[len(context.key_path_block) :]
    arrow = context.key_path[-1][-2:]

    _, key_path = get_target_and_key(context)
    if isinstance(context.input_context, dict):
        nested_set(
            element=context.input_context,
            keys=base_key_path[-(len(base_key_path) - len(context.key_path_block)) :]
            + new_key,
            value={arrow: 'var', 'input': element},
        )
        # Remove the old key
        nested_delete(context.input_context, old_key_path)
        context.key_path = base_key_path + new_key

    else:
        context.input_context = {arrow: 'var', 'input': element}
        context.key_path = base_key_path

    return {arrow: 'var'}


def raise_on_private_hook(k: str, context: 'Context', func_name: str):
    raise exceptions.MalformedFunctionFieldException(
        f"The field {k} can not be a private hook call (ie ending in '_>') as there is "
        f"no situation this makes sense.",
        function_name=func_name,
        context=context,
    )


FUNCTION_ARGS = {
    'help',
    'exec',
    'exec<-',
    'return',
    'args',
    'kwargs',
    'render_exclude',
    'extends',
}


def function_field_to_parseable_macro(
    func_dict: dict, context: 'Context', func_name: str
) -> dict:
    """
    Convert function field to parseable subclassed dict that can be caught later and
     parsed in the event that the value is not supplied via some arg / kwarg.
    """
    for k, v in func_dict.copy().items():
        if k in FUNCTION_ARGS:
            continue

        elif k[-2:] == '->':
            new_value = {'default': {k[-2:]: v}}
            func_dict = {
                key if key != k else k[:-2]: value if key != k else new_value
                for key, value in func_dict.items()
            }  # noqa

        elif k[-2:] == '_>':
            raise_on_private_hook(k, context=context, func_name=func_name)

        elif not isinstance(v, dict):
            continue

        elif '->' in v:
            func_dict[k] = {'default': v}

        elif '_>' in v:
            raise_on_private_hook(k, context=context, func_name=func_name)

        elif 'default->' in v:
            new_value = func_dict[k].pop('default->')
            func_dict[k]['default'] = {'->': new_value}

        elif 'default_>' in v:
            raise_on_private_hook(k, context=context, func_name=func_name)

    return func_dict



import os

from tackle.models import BaseHook, Field


class OsSystemHook(BaseHook):
    """Run system commands."""

    hook_type: str = 'os_system'

    command: str = Field(..., description="A shell command.")
    ignore_error: bool = Field(False, description="Ignore errors.")

    args: list = ['command']

    def exec(self) -> None:
        # try:
        os.system(self.command)
        # except



"""Test for tackle.utils.command"""
import pytest

from tackle.utils.command import unpack_args_kwargs_string, split_input_string

INPUT_STRINGS = [
    ('this --if "expanded == \'that\'"', ['this', '--if', "expanded == 'that'"]),
    ('this that', ['this', 'that']),
    ('this ["foo","bar"]', ['this', ["foo", "bar"]]),
    ('this {"foo":"bar"}', ['this', {"foo": "bar"}]),
    ('{"foo":"bar"} this', [{"foo": "bar"}, 'this']),
    (
        '["this"] --if "expanded == \'that\'"',
        [["this"], '--if', "expanded == \'that\'"],
    ),
    (
        '["this","and","that"] --if "expanded == \'that\'"',
        [["this", "and", "that"], '--if', "expanded == \'that\'"],
    ),
    (
        '{"this":"and"} --if "expanded == \'that\'"',
        [{"this": "and"}, '--if', "expanded == \'that\'"],
    ),
    (
        '{"this":"and","that":"and"} --if "expanded == \'that\'"',
        [{"this": "and", "that": "and"}, '--if', "expanded == \'that\'"],
    ),
    (
        "{'this':'and','that':'and'} --if 'expanded == \"that\"'",
        [{"this": "and", "that": "and"}, '--if', "expanded == \"that\""],
    ),
    (
        "{{get('stuff-and')}}",
        ["{{get(\'stuff-and\')}}"],
    ),
    (
        "{{get('stuff-and',sep='-')}}",
        ["{{get(\'stuff-and\',sep=\'-\')}}"],
    ),
    # TODO: Fix splitting on whitespace unless it is surrounded by special chars
    #  Or just replace with pyparsing ->
    # (
    #     "{{print('things', 'stuff')}}",
    #     ["{{print(\'things\', \'stuff\')}}"],
    # ),
]


@pytest.mark.parametrize("input_string,expected_output", INPUT_STRINGS)
def test_utils_command_split_input_string(input_string, expected_output):
    output = split_input_string(input_string)
    assert output == expected_output


TEMPLATES = [
    # # template, len_args, len_kwargs, len_flags
    ('foo bar baz', 3, 0, 0),
    ('foo --bar baz bing', 2, 1, 0),
    ('foo bar --baz foo', 2, 1, 0),
    ('foo bar --baz foo --bing baz', 2, 2, 0),
    ('foo --bar baz', 1, 1, 0),
    ('foo bar --baz', 2, 0, 1),
    ('foo bar baz --bing', 3, 0, 1),
    ('foo --bar baz --foo', 1, 1, 1),
    ('foo bar --foo bar --bing --baz bling', 2, 2, 1),
    ('foo --bar baz blah --bling', 2, 1, 1),
    ('this --if "expanded == \'that\'"', 1, 1, 0),
    ('["this"] --if "expanded == \'that\'"', 1, 1, 0),
    ('["this"] ["this"] --if "expanded == \'that\'"', 2, 1, 0),
    ('["this"] --for ["this"] --if "expanded == \'that\'"', 1, 2, 0),
    ('"this --if" --if "expanded == \'that\'"', 1, 1, 0),
    ('var {{print("things")}}', 2, 0, 0),
    ('tackle secrets.yaml --if isfile(path_join([cwd,\'secrets.yaml\']))', 2, 1, 0),
    # For inputs with `=` signs -> Need to modify insane regex (ugh)
    # -> or move to peg parser
    # ('foo --bar=baz bing', 2, 1, 0),
    # ('foo bar --baz=foo', 2, 1, 0),
    # ('foo bar --baz=foo --bing=baz', 2, 2, 0),
    # ('foo --bar=baz', 1, 1, 0),
    # ('foo --bar=baz --foo', 1, 1, 1),
    # ('foo bar --foo=bar --bing --baz bling', 2, 2, 1),
    # ('foo --bar=baz blah --bling', 2, 1, 1),
    # ('foo --bar= baz blah --bling', 2, 1, 1),
    # ('foo --bar = baz blah --bling', 2, 1, 1),
]


@pytest.mark.parametrize("template,len_args,len_kwargs,len_flags", TEMPLATES)
def test_unpack_args_kwargs(template, len_args, len_kwargs, len_flags):
    """Validate the count of each input arg/kwarg/flag."""
    args, kwargs, flags = unpack_args_kwargs_string(template)

    assert len_args == len(args)
    assert len_kwargs == len(kwargs)
    assert len_flags == len(flags)


FIXTURES = [
    # input_string, args, kwargs, flags
    ("this --if \"expanded == 'that'\"", ["this"], {"if": "expanded == 'that'"}, []),
    (
        "this that --if \"expanded == 'that'\"",
        ["this", "that"],
        {"if": "expanded == 'that'"},
        [],
    ),
]


@pytest.mark.parametrize("input_string,args,kwargs,flags", FIXTURES)
def test_unpack_input_string(input_string, args, kwargs, flags):
    """Validate expressions for input strings."""
    args_out, kwargs_out, flags_out = unpack_args_kwargs_string(input_string)
    assert args_out == args
    assert kwargs_out == kwargs
    assert flags_out == flags



import os
import sys
import random
import string
import importlib.machinery
import json
from pydantic import BaseModel, Field
import inspect
from typing import List, get_type_hints, Any, Union, Optional

try:
    from typing import _GenericAlias
except ImportError:
    pass

from tackle.utils.paths import listdir_absolute
from tackle.models import BaseHook


class HookDocField(BaseModel):
    """Class for params in a hook."""

    name: str
    required: str
    type: str
    default: Any = ""
    description: str = ""


class HookArgField(BaseModel):
    """Class for args in a hook."""

    argument: str
    type: str


class HookDoc(BaseModel):
    """Class for a hook doc."""

    hook_type: str
    description: str = ""
    properties: List[HookDocField]
    arguments: List[HookArgField]
    return_type: Optional[str]
    return_description: Optional[str]
    hook_file_name: str
    doc_tags: list
    issue_numbers: list
    notes: list
    order: int


class ProviderDocs(BaseModel):
    """Class for a provider hook."""

    name: str = None
    hooks: List[HookDoc] = []
    requirements: list = []
    description: str = None


def hook_type_to_string(type_) -> str:
    """Convert hook ModelField type_ to string."""
    if type_ == Any:
        output = 'any'
    elif isinstance(type_, _GenericAlias):
        if isinstance(type_, List):
            output = 'list'
        else:
            output = 'union'
    else:
        output = type_.__name__
    return output


def get_hook_properties(hook) -> List[HookDocField]:
    """Get the input params for a hook."""
    # Get the base properties like `for` and `if` so they can be ignored
    basehook_properties = BaseHook(hook_type='tmp').__fields__

    output = []
    for k, v in hook.__fields__.items():
        # Ignored properties
        if k in basehook_properties:
            continue

        # Type is dealt with elsewhere
        if k == 'hook_type':
            continue

        # Skip properties ending with "_" as these don't need documentation
        if k.endswith("_"):
            continue

        hook_doc = HookDocField(
            name=k,
            required=v.required,
            type=hook_type_to_string(v.type_),
            default=v.default,
            description=v.field_info.description
            if v.field_info.description is not None
            else "",
        )
        output.append(hook_doc)
    return output


def get_hook_arguments(hook) -> List[HookArgField]:
    """Get the arguments for a hook."""
    output = []

    for i in hook.__fields__['args'].default:
        output.append(
            HookArgField(
                argument=i,
                type=hook_type_to_string(hook.__fields__[i].type_),
            )
        )

    return output


class ProviderDocsHook(BaseHook):
    """Hook for extracting provider metadata for building docs."""

    hook_type: str = "provider_docs"

    # fmt: off
    path: str = Field(".", description="The path to the provider.")
    output: str = Field(".", description="The path to output the docs to.")
    provider: str = Field(None, description="The provider name.")
    hooks_dir = Field("hooks", description="Directory hooks are in.")
    output_schemas: bool = Field(False, description="Output the json schema instead.")
    # fmt: on

    args: list = ['path', 'output']
    _doc_tags: list = ["experimental"]
    _docs_order = 11
    _return_description = (
        "Returns a dictionary with metadata about a provider and "
        "it's hooks or a list of schemas when run with "
        "`output_schemas`."
    )

    def check_python_version(self):
        """Doesn't work on py3.6."""
        if sys.version_info.minor <= 6:
            raise Exception("Can't run provider_docs hook in a py version < 3.7.")

    def exec(self) -> Union[dict, list]:
        """Build the docs."""
        if not self.provider:
            self.provider = os.path.basename(os.path.abspath(self.path))

        requirements = []
        if os.path.exists('requirements.txt'):
            with open('requirements.txt') as f:
                requirements = f.readlines()
                requirements = [line.rstrip() for line in requirements]

        docs = ProviderDocs(
            name=self.provider,
            requirements=requirements,
        )

        # fmt: off
        path = os.path.abspath(os.path.join(self.path, self.hooks_dir))
        # Loop through python files
        importable_files = [i for i in listdir_absolute(path) if
                            i.endswith("py") and not i.endswith('__init__.py')]
        schema_list = []
        for i, f in enumerate(importable_files):
            # Extract out the objects that are derived from the BaseHook
            # We generate a random name because we run this multiple times and getting
            # overlapping pydantic validator config error otherwise.
            random_name = ''.join(
                random.choice(string.ascii_uppercase + string.digits) for _ in range(6))
            loader = importlib.machinery.SourceFileLoader(fullname=random_name,
                                                          path=f)
            module_classes = inspect.getmembers(loader.load_module(), inspect.isclass)
            hooks = [
                i[1]
                for i in module_classes
                if issubclass(i[1], BaseHook) and i[1] != BaseHook
            ]
            # fmt: on

            for h in hooks:
                if h._wip:
                    continue

                return_type = get_type_hints(h.exec)
                if 'return' in return_type:
                    return_type = hook_type_to_string(return_type['return'])
                else:
                    return_type = None

                # Generate the json schema
                # schema = json.loads(h.schema_json(exclude={'default_hook'}))
                schema = json.loads(h.schema_json())
                if self.output_schemas:
                    schema_list.append(schema)
                else:
                    # Generate the docs object
                    hook_doc = HookDoc(
                        properties=get_hook_properties(h),
                        arguments=get_hook_arguments(h),
                        hook_type=inspect.signature(h).parameters['hook_type'].default,
                        # In markdown tables, new lines get messed up so replace with html
                        description=schema['description'].replace("\n\n",
                                                                  "<br />").replace(
                            "\n", "")
                        if 'description' in schema
                        else "",
                        return_type=return_type,
                        return_description=h._return_description,
                        hook_file_name=os.path.basename(f),
                        doc_tags=h._doc_tags,
                        issue_numbers=h._issue_numbers,
                        notes=h._notes,
                        order=h._docs_order,
                    )
                    docs.hooks.append(hook_doc)

        # Sort based on order
        # Sort alphabetically
        docs.hooks = sorted(docs.hooks, key=lambda d: d.hook_type)

        if self.output_schemas:
            return schema_list
        else:
            return docs.dict()



import pathlib
import tempfile

from tackle.models import BaseHook, Field


class MakeDirectoryHook(BaseHook):
    """Hook creating a directory."""

    hook_type: str = 'mkdir'
    path: str = Field(..., description="The path to file or directory")

    args: list = ['path']
    _docs_order = 1

    def exec(self) -> str:
        pathlib.Path(self.path).mkdir(parents=True, exist_ok=True)
        return self.path


class MakeTempDirectoryHook(BaseHook):
    """Hook creating a temporary directory."""

    hook_type: str = 'temp_dir'
    _docs_order = 2

    def exec(self) -> str:
        return tempfile.mkdtemp()



"""Tests dict rendering special variables."""
from tackle.main import tackle


def test_special_variables(change_curdir_fixtures):
    """Verify Jinja2 time extension work correctly."""
    output = tackle('special-variables.yaml')
    assert len(output) > 8
